import logging
import os
import os.path as op
import shutil
from copy import copy

import pandas as pd
import requests
import ssbio.protein.structure.properties.msms
import ssbio.protein.structure.properties.quality
import ssbio.protein.structure.properties.residues
from Bio import SeqIO
from Bio.PDB.PDBExceptions import PDBException
from bioservices import KEGG
from bioservices import UniProt
from cobra.core import DictList
from slugify import Slugify
from six.moves.urllib.error import HTTPError

import ssbio.cobra.utils
import ssbio.databases.kegg
import ssbio.databases.pdb
import ssbio.databases.uniprot
import ssbio.protein.sequence.properties.residues
import ssbio.protein.sequence.properties.tmhmm
import ssbio.protein.sequence.utils.fasta
from ssbio import utils
from ssbio.core.genepro import GenePro
from ssbio.core.object import Object
from ssbio.databases.kegg import KEGGProp
from ssbio.databases.uniprot import UniProtProp
from ssbio.protein.sequence.properties.scratch import SCRATCH
from ssbio.protein.structure.homology.itasser.itasserprep import ITASSERPrep

if utils.is_ipynb():
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm

custom_slugify = Slugify(safe_chars='-_.')
logging.getLogger("requests").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)
log = logging.getLogger(__name__)
date = utils.Date()
bs_unip = UniProt()
bs_kegg = KEGG()


class GEMPRO(Object):
    """Generic class to represent all information for a GEM-PRO project.

    Initialize the GEM-PRO project with a genome-scale model, a list of genes, or a dict of genes and sequences.
    Specify the name of your project, along with the root directory where a folder with that name will be created.

    Main methods provided are:

    #. Mapping of sequence IDs
        * With KEGG mapper
        * With UniProt mapper
        * Allowing manual gene ID --> protein sequence entry
        * Allowing manual gene ID --> UniProt ID

    #. Consolidating sequence IDs and setting a representative sequence
        * Currently these are set based on available PDB IDs

    #. Mapping of representative sequence --> structures
        * With UniProt --> ranking of PDB structures
        * BLAST representative sequence --> PDB database

    #. Homology modeling
        * Mapping to existing models
        * Preparation for running I-TASSER
        * Parsing I-TASSER runs

    #. Running QC/QA on structures and setting a representative structure
        * Various cutoffs (mutations, insertions, deletions) can be set to filter structures

    Each step will also generate reports as Pandas DataFrames and also output various logging messages.

    Args:
        gem_name (str): The name of your GEM or just your project in general.
            This will be the name of the main folder that is created in root_dir.
        pdb_file_type (str): pdb, pdb.gz, mmcif, cif, cif.gz, xml.gz, mmtf, mmtf.gz - Choose a file type for
            files downloaded from the PDB
        create_dirs (bool): If GEM-PRO directories should be created - if not,
            most mapping functions need an output directory specified.
        root_dir (str): Path to where the folder named gem_name will be created.
            Default is current working directory.
        genome_path (str): Simple reference link to the genome FASTA file (CDS)
        gem (Model): COBRApy Model object
        gem_file_path (str): Path to GEM file
        gem_file_type (str): GEM model type - 'sbml' (or 'xml'), 'mat', or 'json' formats
        genes_list (list): List of gene IDs that you want to map
        genes_and_sequences (dict): Dictionary of gene IDs and their amino acid sequence strings
        description (str): Optional string to describe your project

    """

    def __init__(self, gem_name, pdb_file_type='cif', create_dirs=True, root_dir=None, genome_path=None,
                 gem=None,
                 gem_file_path=None, gem_file_type=None,
                 genes_list=None,
                 genes_and_sequences=None,
                 description=None):
        Object.__init__(self, id=gem_name, description=description)

        self.pdb_file_type = pdb_file_type
        self.genome_path = genome_path

        self.model = None
        # Load a model object
        if gem:
            self._load_cobra_model(gem)

        # Or, load a GEM file
        elif gem_file_path and gem_file_type:
            gem_manual = ssbio.cobra.utils.model_loader(gem_file_path, gem_file_type, pdb_file_type)
            self._load_cobra_model(gem_manual)

        # Or, load a list of gene IDs
        elif genes_list:
            self.genes = genes_list

        # Or, load a dictionary of genes and their sequences
        elif genes_and_sequences:
            self.genes = list(genes_and_sequences.keys())

        # If neither a model or genes are input, you can still add IDs with method add_genes_by_id later
        else:
            self.genes = []
            log.warning('No model or genes input')

        # Create directories
        self._root_dir = None
        if create_dirs:
            if not root_dir:
                root_dir = os.getcwd()
            self.root_dir = root_dir
        if genes_and_sequences:
            self.manual_seq_mapping(genes_and_sequences)

        log.info('{}: number of genes'.format(len(self.genes)))

    @property
    def root_dir(self):
        """str: Directory where GEM-PRO project folder (base_dir) is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        self._root_dir = path

        if path:
            for d in [self.base_dir, self.model_dir, self.data_dir, self.genes_dir]:
                ssbio.utils.make_dir(d)

            log.info('{}: GEM-PRO project location'.format(self.base_dir))

            # Additionally set the gene folder paths
            for g in self.genes:
                g.root_dir = self.genes_dir
        else:
            log.warning('Root directory not set')

    @property
    def base_dir(self):
        """str: GEM-PRO project folder"""
        if self.root_dir:
            return op.join(self.root_dir, self.id)
        else:
            log.warning('Root directory not set')
            return None

    @property
    def model_dir(self):
        """str: Directory where original GEMs and GEM-related files are stored"""
        if self.root_dir:
            return op.join(self.base_dir, 'model')
        else:
            log.warning('Root directory not set')
            return None

    @property
    def data_dir(self):
        """str: Directory where all data (dataframes and more) will be stored"""
        if self.root_dir:
            return op.join(self.base_dir, 'data')
        else:
            log.warning('Root directory not set')
            return None

    @property
    def genes_dir(self):
        """str: Directory where all gene specific information is stored"""
        if self.root_dir:
            return op.join(self.base_dir, 'genes')
        else:
            log.warning('Root directory not set')
            return None

    def _load_cobra_model(self, model):
        """Load a COBRApy Model object into the GEM-PRO project.

        Args:
            model (Model): COBRApy Model object

        """
        self.model = model
        self.genes = model.genes
        log.info('{}: loaded model'.format(model.id))
        log.info('{}: number of reactions'.format(len(self.model.reactions)))
        log.info('{}: number of reactions linked to a gene'.format(ssbio.cobra.utils.true_num_reactions(self.model)))
        log.info('{}: number of genes (excluding spontaneous)'.format(ssbio.cobra.utils.true_num_genes(self.model)))
        log.info('{}: number of metabolites'.format(len(self.model.metabolites)))
        log.warning('IMPORTANT: All Gene objects have been transformed into GenePro objects, and will be for any new ones')

    @property
    def genes_with_structures(self):
        """Return a DictList of all genes with any available protein structures"""
        return DictList(x for x in self.genes if x.protein.num_structures > 0)

    @property
    def genes_with_experimental_structures(self):
        """Return a DictList of all genes that have at least one experimental structure"""
        return DictList(x for x in self.genes_with_structures if x.protein.num_structures_experimental > 0)

    @property
    def genes_with_homology_models(self):
        """Return a DictList of all genes that have at least one homology model"""
        return DictList(x for x in self.genes_with_structures if x.protein.num_structures_homology > 0)

    @property
    def genes_with_a_representative_sequence(self):
        """Return a DictList of all genes with a representative sequence"""
        tmp = DictList(x for x in self.genes if x.protein.representative_sequence)
        return DictList(y for y in tmp if y.protein.representative_sequence.sequence_file)

    @property
    def genes_with_a_representative_structure(self):
        """Return a DictList of all genes with a representative protein structure"""
        tmp = DictList(x for x in self.genes if x.protein.representative_structure)
        return DictList(y for y in tmp if y.protein.representative_structure.structure_file)

    @property
    def genes(self):
        return ssbio.cobra.utils.filter_out_spontaneous_genes(self._genes)

    @genes.setter
    def genes(self, genes_list):
        """Set the genes attribute to be a DictList of GenePro objects.

        A "protein" attribute will be added to each Gene.

        Args:
            genes_list: DictList of COBRApy Gene objects, or list of gene IDs

        """

        if not isinstance(genes_list, DictList):
            tmp_list = []
            for x in list(set(genes_list)):
                x = str(x)
                new_gene = GenePro(id=x, pdb_file_type=self.pdb_file_type)
                tmp_list.append(new_gene)
            self._genes = DictList(tmp_list)
        else:
            self._genes = genes_list

    def add_genes_by_id(self, genes_list):
        """Add gene IDs manually into our GEM-PRO project.

        Args:
            genes_list (list): List of gene IDs as strings.

        """

        new_genes = []
        for x in list(set(genes_list)):
            new_gene = GenePro(id=x, pdb_file_type=self.pdb_file_type)
            new_genes.append(new_gene)

        # Add unique genes only
        self.genes.union(new_genes)

    ####################################################################################################################
    ### SEQUENCE RELATED METHODS ###
    def kegg_mapping_and_metadata(self, kegg_organism_code, custom_gene_mapping=None, outdir=None,
                                  set_as_representative=False, force_rerun=False):
        """Map all genes in the model to KEGG IDs using the KEGG service.

        Steps:
            1. Download all metadata and sequence files in the sequences directory
            2. Creates a KEGGProp object in the protein.sequences attribute
            3. Returns a Pandas DataFrame of mapping results

        Args:
            kegg_organism_code (str): The three letter KEGG code of your organism
            custom_gene_mapping (dict): If your model genes differ from the gene IDs you want to map,
                custom_gene_mapping allows you to input a dictionary which maps model gene IDs to new ones.
                Dictionary keys must match model gene IDs.
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            set_as_representative (bool): If mapped KEGG IDs should be set as representative sequences
            force_rerun (bool): If you want to overwrite any existing mappings and files

        """

        # First map all of the organism's KEGG genes to UniProt
        kegg_to_uniprot = ssbio.databases.kegg.map_kegg_all_genes(organism_code=kegg_organism_code, target_db='uniprot')

        successfully_mapped_counter = 0

        for g in tqdm(self.genes):
            if custom_gene_mapping:
                kegg_g = custom_gene_mapping[g.id]
            else:
                kegg_g = g.id

            # Download both FASTA and KEGG metadata files
            kegg_prop = g.protein.load_kegg(kegg_id=kegg_g, kegg_organism_code=kegg_organism_code,
                                            download=True, outdir=outdir, set_as_representative=set_as_representative,
                                            force_rerun=force_rerun)

            # Update potentially old UniProt ID
            if kegg_g in kegg_to_uniprot.keys():
                kegg_prop.uniprot = kegg_to_uniprot[kegg_g]
                if g.protein.representative_sequence:
                    if g.protein.representative_sequence.kegg == kegg_prop.kegg:
                        g.protein.representative_sequence.uniprot = kegg_to_uniprot[kegg_g]

            # Keep track of missing mappings - missing is defined by no available sequence
            if kegg_prop.sequence_file:
                successfully_mapped_counter += 1

            log.debug('{}: loaded KEGG information for gene'.format(g.id))

        log.info('{}/{}: number of genes mapped to KEGG'.format(successfully_mapped_counter, len(self.genes)))
        log.info('Completed ID mapping --> KEGG. See the "df_kegg_metadata" attribute for a summary dataframe.')

    @property
    def df_kegg_metadata(self):
        kegg_pre_df = []
        df_cols = ['gene', 'kegg', 'refseq', 'uniprot', 'num_pdbs', 'pdbs', 'seq_len', 'sequence_file', 'metadata_file']

        for g in self.genes:
            kegg_mappings = g.protein.filter_sequences(KEGGProp)
            for kegg_prop in kegg_mappings:
                kegg_dict = kegg_prop.get_dict(df_format=True, only_keys=df_cols)
                kegg_dict['gene'] = g.id
                kegg_pre_df.append(kegg_dict)

        # Save a dataframe of the file mapping info
        df = pd.DataFrame.from_records(kegg_pre_df, columns=df_cols).set_index('gene')
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df)

    @property
    def missing_kegg_mapping(self):
        kegg_missing = []
        for g in self.genes:
            keggs = g.protein.filter_sequences(KEGGProp)
            no_sequence_file_available = True
            for k in keggs:
                if k.sequence_file:
                    no_sequence_file_available = False
                    break
            if no_sequence_file_available:
                kegg_missing.append(g.id)
        return list(set(kegg_missing))

    def uniprot_mapping_and_metadata(self, model_gene_source, custom_gene_mapping=None, outdir=None,
                                     set_as_representative=False, force_rerun=False):
        """Map all genes in the model to UniProt IDs using the UniProt mapping service.
        Also download all metadata and sequences.

        Args:
            model_gene_source (str): the database source of your model gene IDs.
                See: http://www.uniprot.org/help/programmatic_access.
                Common model gene sources are:
                    * Ensembl Genomes - ``ENSEMBLGENOME_ID`` (i.e. E. coli b-numbers)
                    * Entrez Gene (GeneID) - ``P_ENTREZGENEID``
                    * RefSeq Protein - ``P_REFSEQ_AC``
            custom_gene_mapping (dict): If your model genes differ from the gene IDs you want to map,
                custom_gene_mapping allows you to input a dictionary which maps model gene IDs to new ones.
                Dictionary keys must match model genes.
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            set_as_representative (bool): If mapped UniProt IDs should be set as representative sequences
            force_rerun (bool): If you want to overwrite any existing mappings and files

        """

        # Allow model gene --> custom ID mapping ({'TM_1012':'TM1012'})
        if custom_gene_mapping:
            genes_to_map = list(custom_gene_mapping.values())
        else:
            genes_to_map = [x.id for x in self.genes]

        # Map all IDs first to available UniProts
        genes_to_uniprots = bs_unip.mapping(fr=model_gene_source, to='ACC', query=genes_to_map)

        successfully_mapped_counter = 0
        for g in tqdm(self.genes):
            if custom_gene_mapping and g.id in custom_gene_mapping.keys():
                uniprot_gene = custom_gene_mapping[g.id]
            else:
                uniprot_gene = g.id

            if uniprot_gene not in list(genes_to_uniprots.keys()):
                log.debug('{}: unable to map to UniProt'.format(g.id))
            else:
                for mapped_uniprot in genes_to_uniprots[uniprot_gene]:
                    try:
                        uniprot_prop = g.protein.load_uniprot(uniprot_id=mapped_uniprot, download=True, outdir=outdir,
                                                              set_as_representative=set_as_representative,
                                                              force_rerun=force_rerun)
                    except HTTPError as e:
                        print(e)
                        log.error('{}, {}: unable to complete web request'.format(g.id, mapped_uniprot))
                        continue

                    if uniprot_prop.sequence_file:
                        successfully_mapped_counter += 1

        log.info('{}/{}: number of genes mapped to UniProt'.format(successfully_mapped_counter, len(self.genes)))
        log.info('Completed ID mapping --> UniProt. See the "df_uniprot_metadata" attribute for a summary dataframe.')

    def manual_uniprot_mapping(self, gene_to_uniprot_dict, outdir=None, set_as_representative=True):
        """Read a manual dictionary of model gene IDs --> UniProt IDs. By default sets them as representative.

        This allows for mapping of the missing genes, or overriding of automatic mappings.

        Args:
            gene_to_uniprot_dict: Dictionary of mappings with key:value pairs:
                <gene_id1>:<uniprot_id1>,
                <gene_id2>:<uniprot_id2>,
                ...
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            set_as_representative (bool): If mapped UniProt IDs should be set as representative sequences

        """
        for g, u in gene_to_uniprot_dict.items():
            g = str(g)
            gene = self.genes.get_by_id(g)

            try:
                uniprot_prop = gene.protein.load_uniprot(uniprot_id=u,
                                                         outdir=outdir, download=True,
                                                         set_as_representative=set_as_representative)
            except requests.HTTPError as e:
                print(e)
                log.error('{}, {}: unable to complete web request'.format(g, u))
                continue

        log.info('Completed manual ID mapping --> UniProt. See the "df_uniprot_metadata" attribute for a summary dataframe.')

    @property
    def df_uniprot_metadata(self):
        uniprot_pre_df = []
        df_cols = ['gene', 'uniprot', 'reviewed', 'gene_name', 'kegg', 'refseq', 'num_pdbs', 'pdbs', 'ec_number',
                   'pfam', 'seq_len', 'description', 'entry_version', 'seq_version', 'sequence_file', 'metadata_file']

        for g in self.genes:
            uniprot_mappings = g.protein.filter_sequences(UniProtProp)
            for uniprot_prop in uniprot_mappings:
                uniprot_dict = uniprot_prop.get_dict(df_format=True, only_keys=df_cols)
                uniprot_dict['gene'] = g
                uniprot_pre_df.append(uniprot_dict)

        df = pd.DataFrame.from_records(uniprot_pre_df, columns=df_cols).set_index('gene')
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df)

    @property
    def missing_uniprot_mapping(self):
        uniprot_missing = []
        for g in self.genes:
            ups = g.protein.filter_sequences(UniProtProp)

            no_sequence_file_available = True
            for u in ups:
                if u.sequence_file:
                    no_sequence_file_available = False
                    break
            if no_sequence_file_available:
                uniprot_missing.append(g.id)

        return list(set(uniprot_missing))

    # TODO: should also have a seq --> uniprot id function (has to be 100% match) (also needs organism)
    def manual_seq_mapping(self, gene_to_seq_dict, outdir=None, set_as_representative=True):
        """Read a manual input dictionary of model gene IDs --> protein sequences. By default sets them as representative.

        Args:
            gene_to_seq_dict (dict): Mapping of gene IDs to their protein sequence strings
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            set_as_representative (bool): If mapped sequences should be set as representative

        """
        if outdir:
            outdir_set = True
        else:
            outdir_set = False

        # Save the sequence information in individual FASTA files
        for g, s in gene_to_seq_dict.items():
            g = str(g)
            gene = self.genes.get_by_id(g)

            if not outdir_set:
                outdir = gene.protein.sequence_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')

            manual_info = gene.protein.load_manual_sequence(ident=g, seq=s, outdir=outdir,
                                                            write_fasta_file=True,
                                                            set_as_representative=set_as_representative)
            log.debug('{}: loaded manually defined sequence information'.format(g))

        log.info('Loaded in {} sequences'.format(len(gene_to_seq_dict)))

    def set_representative_sequence(self, force_rerun=False):
        """Combine information from KEGG, UniProt, and manual mappings. Saves a DataFrame of results.

            Manual mappings override all existing mappings. UniProt mappings override KEGG mappings except
            when KEGG mappings have PDBs associated with them and UniProt doesn't.

        """

        # TODO: rethink use of multiple database sources - may lead to inconsistency with genome sources

        sequence_missing = []
        successfully_mapped_counter = 0
        for g in tqdm(self.genes):
            repseq = g.protein.set_representative_sequence(force_rerun=force_rerun)

            if not repseq:
                sequence_missing.append(g.id)
            elif not repseq.sequence_path:
                sequence_missing.append(g.id)
            else:
                successfully_mapped_counter += 1

        log.info('{}/{}: number of genes with a representative sequence'.format(len(self.genes_with_a_representative_sequence),
                                                                                len(self.genes)))
        log.info('See the "df_representative_sequences" attribute for a summary dataframe.')

    @property
    def df_representative_sequences(self):
        seq_mapping_pre_df = []
        df_cols = ['gene', 'uniprot', 'kegg', 'num_pdbs', 'pdbs', 'seq_len', 'sequence_file', 'metadata_file']

        for g in self.genes_with_a_representative_sequence:
            gene_dict = g.protein.representative_sequence.get_dict(df_format=True, only_keys=df_cols)
            gene_dict['gene'] = g.id
            seq_mapping_pre_df.append(gene_dict)

        df = pd.DataFrame.from_records(seq_mapping_pre_df, columns=df_cols).set_index('gene')
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df)

    @property
    def missing_representative_sequence(self):
        return [x.id for x in self.genes if not self.genes_with_a_representative_sequence.has_id(x.id)]

    def write_representative_sequences_file(self, outname, outdir=None, set_ids_from_model=True):
        """Write all the model's sequences as a single FASTA file. By default, sets IDs to model gene IDs.

        Args:
            outname (str): Name of the output FASTA file without the extension
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            set_ids_from_model (bool): If the gene ID source should be the model gene IDs, not the original sequence ID

        """

        if not outdir:
            outdir = self.data_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        outfile = op.join(outdir, outname + '.faa')

        tmp = []
        for x in self.genes_with_a_representative_sequence:
            repseq = x.protein.representative_sequence
            copied_seq_record = copy(repseq.seq_record)
            if set_ids_from_model:
                copied_seq_record.id = x.id
            tmp.append(copied_seq_record)

        SeqIO.write(tmp, outfile, "fasta")

        log.info('{}: wrote all representative sequences to file'.format(outfile))
        self.genome_path = outfile
        return self.genome_path

    def get_sequence_properties(self):
        """Run Biopython ProteinAnalysis and EMBOSS pepstats to summarize basic statistics of the protein sequences.

        Annotations are stored in the gene protein's representative sequence at .seq_record.annotations

        """
        for g in tqdm(self.genes_with_a_representative_sequence):
            g.protein.representative_sequence.get_biopython_pepstats()
            g.protein.representative_sequence.get_emboss_pepstats()

    def get_scratch_predictions(self, path_to_scratch, results_dir, scratch_basename='scratch', num_cores=1,
                                exposed_buried_cutoff=25, custom_gene_mapping=None):
        """Run and parse SCRATCH results to predict secondary structure and solvent accessibility.

        Annotations are stored in the gene protein's representative sequence at:
            * .seq_record.annotations
            * .seq_record.letter_annotations

        Args:
            path_to_scratch (str): Path to SCRATCH executable
            results_dir (str): Path to SCRATCH results folder, which will have the files (scratch.ss, scratch.ss8,
                scratch.acc, scratch.acc20)
            scratch_basename (str): Basename of the SCRATCH results ('scratch' is default)
            exposed_buried_cutoff (int): Cutoff of exposed/buried for the acc20 predictions
            custom_gene_mapping (dict): Default parsing of SCRATCH output files is to look for the model gene IDs. If
                your output files contain IDs which differ from the model gene IDs, use this dictionary to map model
                gene IDs to result file IDs. Dictionary keys must match model genes.

        """
        if not self.genome_path:
            # Write all sequences as one file
            all_seqs = self.write_representative_sequences_file(outname='{}_cds'.format(self.id))

        # Runs SCRATCH or loads existing results in results_dir
        scratch = SCRATCH(project_name=scratch_basename, seq_file=self.genome_path)
        scratch.run_scratch(path_to_scratch=path_to_scratch, num_cores=num_cores, outdir=results_dir)

        counter = 0

        # Adding the scratch annotations to the representative_sequences letter_annotations
        for g in tqdm(self.genes_with_a_representative_sequence):
            if custom_gene_mapping:
                g_id = custom_gene_mapping[g.id]
            else:
                g_id = g.id

            if g_id in scratch.sspro_summary():
                # Secondary structure
                g.protein.representative_sequence.seq_record.annotations.update(scratch.sspro_summary()[g_id])
                g.protein.representative_sequence.seq_record.annotations.update(scratch.sspro8_summary()[g_id])
                g.protein.representative_sequence.load_letter_annotations('SS-sspro', scratch.sspro_results()[g_id])
                g.protein.representative_sequence.load_letter_annotations('SS-sspro8', scratch.sspro8_results()[g_id])

                # Solvent accessibility
                g.protein.representative_sequence.seq_record.annotations.update(scratch.accpro_summary()[g_id])
                g.protein.representative_sequence.seq_record.annotations.update(scratch.accpro20_summary(exposed_buried_cutoff)[g_id])
                g.protein.representative_sequence.load_letter_annotations('RSA-accpro', scratch.accpro_results()[g_id])
                g.protein.representative_sequence.load_letter_annotations('RSA-accpro20', scratch.accpro20_results()[g_id])

                counter += 1
            else:
                log.error('{}: missing SCRATCH results'.format(g.id))

        log.info('{}/{}: number of genes with SCRATCH predictions loaded'.format(counter, len(self.genes)))

    def get_tmhmm_predictions(self, tmhmm_results, custom_gene_mapping=None):
        """Parse TMHMM results and store in the representative sequences

        Args:
            tmhmm_results (str): Path to TMHMM results (long format)
            custom_gene_mapping (dict): Default parsing of TMHMM output is to look for the model gene IDs. If
                your output file contains IDs which differ from the model gene IDs, use this dictionary to map model
                gene IDs to result file IDs. Dictionary keys must match model genes.

        """
        tmhmm_dict = ssbio.protein.sequence.properties.tmhmm.parse_tmhmm_long(tmhmm_results)

        counter = 0
        for g in tqdm(self.genes_with_a_representative_sequence):
            if custom_gene_mapping:
                g_id = custom_gene_mapping[g.id]
            else:
                g_id = g.id

            if g_id in tmhmm_dict:
                log.debug('{}: loading TMHMM results'.format(g.id))
                if not tmhmm_dict[g_id]:
                    log.error("{}: missing TMHMM results".format(g.id))
                g.protein.representative_sequence.seq_record.annotations['num_tm_helix-tmhmm'] = tmhmm_dict[g_id]['num_tm_helices']
                g.protein.representative_sequence.load_letter_annotations('TM-tmhmm', tmhmm_dict[g_id]['sequence'])
                counter += 1
            else:
                log.error("{}: missing TMHMM results".format(g.id))

        log.info('{}/{}: number of genes with TMHMM predictions loaded'.format(counter, len(self.genes)))

    ### END SEQUENCE RELATED METHODS ###
    ####################################################################################################################

    ####################################################################################################################
    ### STRUCTURE RELATED METHODS ###
    def map_uniprot_to_pdb(self, seq_ident_cutoff=0.0, outdir=None, force_rerun=False):
        """Map UniProt IDs to a ranked list of PDB structures available.

        Uses the "Best structures" API availble from https://www.ebi.ac.uk/pdbe/api/doc/sifts.html
        The list of PDB structures mapping to a UniProt accession sorted by coverage of the protein and,
        if the same, resolution. Also creates a summary dataframe accessible by the attribute "df_pdb_ranking".

        Args:
            seq_ident_cutoff (float): Cutoff results based on percent coverage (in decimal form)
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            force_rerun (bool): Obtain best structures mapping ignoring previously downloaded results

        """

        # First get all UniProt IDs and check if they have PDBs
        all_representative_uniprots = []
        for g in self.genes_with_a_representative_sequence:
            uniprot_id = g.protein.representative_sequence.uniprot
            if uniprot_id:
                all_representative_uniprots.append(uniprot_id)
        log.info('Mapping UniProt IDs --> PDB IDs...')
        uniprots_to_pdbs = bs_unip.mapping(fr='ACC', to='PDB_ID', query=all_representative_uniprots)

        counter = 0
        # Now run the best_structures API for all genes
        for g in tqdm(self.genes_with_a_representative_sequence):
            uniprot_id = g.protein.representative_sequence.uniprot
            if uniprot_id:
                if uniprot_id in uniprots_to_pdbs:
                    best_structures = g.protein.map_uniprot_to_pdb(seq_ident_cutoff=seq_ident_cutoff, outdir=outdir, force_rerun=force_rerun)
                    if best_structures:
                        counter += 1
                        log.debug('{}: {} PDBs mapped'.format(g.id, len(best_structures)))
                else:
                    log.debug('{}, {}: no PDBs available'.format(g.id, uniprot_id))

        log.info('{}/{}: number of genes with at least one experimental structure'.format(len(self.genes_with_experimental_structures),
                                                                                          len(self.genes)))
        log.info('Completed UniProt --> best PDB mapping. See the "df_pdb_ranking" attribute for a summary dataframe.')

    @property
    def df_pdb_ranking(self):
        df = pd.DataFrame()
        for g in self.genes_with_experimental_structures:
            protein_df = g.protein.df_pdb_ranking.copy().reset_index()
            if not protein_df.empty:
                protein_df['gene'] = g.id
                df = df.append(protein_df)
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df.set_index('gene'))

    def blast_seqs_to_pdb(self, seq_ident_cutoff=0, evalue=0.0001, all_genes=False, display_link=False,
                          outdir=None, force_rerun=False):
        """BLAST each gene sequence to the PDB. Saves raw BLAST results (XML files).

        Also creates a summary dataframe accessible by the attribute "df_pdb_blast".

        Args:
            seq_ident_cutoff (float, optional): Cutoff results based on percent coverage (in decimal form)
            evalue (float, optional): Cutoff for the E-value - filters for significant hits. 0.001 is liberal, 0.0001 is stringent (default).
            all_genes (bool): If all genes should be BLASTed, or only those without any structures currently mapped
            display_link (bool, optional): Set to True if links to the HTML results should be displayed
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            force_rerun (bool, optional): If existing BLAST results should not be used, set to True. Default is False

        """
        counter = 0

        for g in tqdm(self.genes):
            # If all_genes=False, BLAST only genes without a uniprot -> pdb mapping
            if g.protein.num_structures > 0 and not all_genes and not force_rerun:
                log.debug('{}: skipping BLAST, {} structures already mapped and all_genes flag is False'.format(g.id,
                                                                                                                g.protein.num_structures))
                continue

            # BLAST the sequence to the PDB
            new_pdbs = g.protein.blast_representative_sequence_to_pdb(seq_ident_cutoff=seq_ident_cutoff,
                                                                   evalue=evalue,
                                                                   display_link=display_link,
                                                                   outdir=outdir,
                                                                   force_rerun=force_rerun)

            if new_pdbs:
                counter += 1
                log.debug('{}: {} PDBs BLASTed'.format(g.id, len(new_pdbs)))
            else:
                log.debug('{}: no BLAST results'.format(g.id))

        log.info('Completed sequence --> PDB BLAST. See the "df_pdb_blast" attribute for a summary dataframe.')
        log.info('{}: number of genes with additional structures added from BLAST'.format(counter))

    @property
    def df_pdb_blast(self):
        df = pd.DataFrame()
        for g in self.genes_with_experimental_structures:
            protein_df = g.protein.df_pdb_blast.copy().reset_index()
            if not protein_df.empty:
                protein_df['gene'] = g.id
                df = df.append(protein_df)
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df.set_index('gene'))

    def manual_homology_models(self, input_dict, outdir=None, clean=True, force_rerun=False):
        """Copy homology models to the GEM-PRO project.

        Args:
            input_dict: Dictionary of dictionaries of gene names to homology model IDs and information. Input a dict of:
                {model_gene: {homology_model_id1: {'model_file': '/path/to/homology/model',
                                                   'file_type': 'pdb'},
                              homology_model_id2: {'model_file': '/path/to/homology/model'
                                                   'file_type': 'pdb'}
                             },
                }
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            clean (bool): If homology files should be cleaned and saved as a new PDB file
            force_rerun (bool): If homology files should be copied again even if they exist in the GEM-PRO directory

        """
        if outdir:
            outdir_set = True
        else:
            outdir_set = False

        counter = 0
        for g in tqdm(self.genes):
            if g.id not in input_dict:
                continue

            if not outdir_set:
                outdir = g.protein.structure_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')

            for hid, hdict in input_dict[g.id].items():
                if 'model_file' not in hdict or 'file_type' not in hdict:
                    raise KeyError('"model_file" and "file_type" must be keys in the manual input dictionary.')

                new_homology = g.protein.load_generic_structure(hid, hdict['model_file'])

                if clean:
                    new_homology.load_structure_path(new_homology.clean_structure(outdir=outdir, force_rerun=force_rerun),
                                                     hdict['file_type'])
                else:
                    if ssbio.utils.force_rerun(force_rerun, op.basename(hdict['model_file'])):
                        # Just copy the file to the structure directory and store the file name
                        shutil.copy2(hdict['model_file'], outdir)
                        new_homology.load_structure_path(op.join(outdir, hdict['model_file']), hdict['file_type'])

                # TODO: need to handle other info in the provided dictionary, if any
                new_homology.update(hdict)

                log.debug('{}: updated homology model information and copied model file.'.format(g.id))
            counter += 1

        log.info('Updated homology model information for {} genes.'.format(counter))

    def get_itasser_models(self, homology_raw_dir, custom_itasser_name_mapping=None, outdir=None, force_rerun=False):
        """Copy generated I-TASSER models from a directory to the GEM-PRO directory.

        Args:
            homology_raw_dir (str): Root directory of I-TASSER folders.
            custom_itasser_name_mapping (dict): Use this if your I-TASSER folder names differ from your model gene names.
                Input a dict of {model_gene: ITASSER_folder}.
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            force_rerun (bool): If homology files should be copied again even if they exist in the GEM-PRO directory

        """
        counter = 0

        for g in tqdm(self.genes):
            if custom_itasser_name_mapping and g.id in custom_itasser_name_mapping:
                hom_id = custom_itasser_name_mapping[g.id]
                if not op.exists(op.join(homology_raw_dir, hom_id)):
                    hom_id = g.id
            else:
                hom_id = g.id

            # The name of the actual pdb file will be $GENEID_model1.pdb
            new_itasser_name = hom_id + '_model1'
            orig_itasser_dir = op.join(homology_raw_dir, hom_id)

            try:
                itasser_prop = g.protein.load_itasser_folder(ident=hom_id, itasser_folder=orig_itasser_dir,
                                                             organize=True, outdir=outdir, organize_name=new_itasser_name,
                                                             create_dfs=True, force_rerun=force_rerun)
            except OSError:
                log.debug('{}: homology model folder unavailable'.format(g.id))
                continue

            if itasser_prop.model_file:
                counter += 1
            else:
                log.debug('{}: homology model file unavailable, perhaps modelling did not finish'.format(g.id))

        log.info('Completed copying of {} I-TASSER models to GEM-PRO directory. See the "df_homology_models" attribute for a summary dataframe.'.format(counter))

    @property
    def df_homology_models(self):
        df = pd.DataFrame()
        for g in self.genes_with_homology_models:
            protein_df = g.protein.df_homology_models.copy().reset_index()
            if not protein_df.empty:
                protein_df['gene'] = g.id
                df = df.append(protein_df)
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df.set_index('gene'))

    def set_representative_structure(self, seq_outdir=None, struct_outdir=None, pdb_file_type=None,
                                     engine='needle', always_use_homology=False,
                                     seq_ident_cutoff=0.5, allow_missing_on_termini=0.2,
                                     allow_mutants=True, allow_deletions=False,
                                     allow_insertions=False, allow_unresolved=True,
                                     force_rerun=False):
        """Set the representative protein structure for a gene.

        Each gene can have a combination of the following, which will be analyzed to set a representative structure.
            * Homology model(s)
            * Ranked PDBs
            * BLASTed PDBs

        If the always_use_homology flag is true, homology models are always set as representative when they exist.
        If there are multiple homology models, we rank by the percent sequence coverage.

        Args:
            seq_outdir (str): Path to output directory of sequence alignment files, must be set if GEM-PRO directories
                were not created initially
            struct_outdir (str): Path to output directory of structure files, must be set if GEM-PRO directories
                were not created initially
            pdb_file_type (str): ``pdb``, ``pdb.gz``, ``mmcif``, ``cif``, ``cif.gz``, ``xml.gz``, ``mmtf``,
                ``mmtf.gz`` - PDB structure file type that should be downloaded
            engine (str): ``needle`` or ``biopython`` - which pairwise sequence alignment engine should be used
                needle is the standard EMBOSS tool to run pairwise alignments
                biopython is Biopython's implementation of needle. Results can differ!
            always_use_homology (bool): If homology models should always be set as the representative structure
            seq_ident_cutoff (float): Percent sequence identity cutoff, in decimal form
            allow_missing_on_termini (float): Percentage of the total length of the reference sequence which will be ignored
                when checking for modifications. Example: if 0.1, and reference sequence is 100 AA, then only residues
                5 to 95 will be checked for modifications.
            allow_mutants (bool): If mutations should be allowed or checked for
            allow_deletions (bool): If deletions should be allowed or checked for
            allow_insertions (bool): If insertions should be allowed or checked for
            allow_unresolved (bool): If unresolved residues should be allowed or checked for
            force_rerun (bool): If sequence to structure alignment should be rerun

        """
        for g in tqdm(self.genes):
            repstruct = g.protein.set_representative_structure(seq_outdir=seq_outdir,
                                                               struct_outdir=struct_outdir,
                                                               pdb_file_type=pdb_file_type,
                                                               engine=engine,
                                                               seq_ident_cutoff=seq_ident_cutoff,
                                                               always_use_homology=always_use_homology,
                                                               allow_missing_on_termini=allow_missing_on_termini,
                                                               allow_mutants=allow_mutants,
                                                               allow_deletions=allow_deletions,
                                                               allow_insertions=allow_insertions,
                                                               allow_unresolved=allow_unresolved,
                                                               force_rerun=force_rerun)

        log.info('{}/{}: number of genes with a representative structure'.format(len(self.genes_with_a_representative_structure),
                                                                                 len(self.genes)))
        log.info('See the "df_representative_structures" attribute for a summary dataframe.')

    @property
    def df_representative_structures(self):
        rep_struct_pre_df = []
        df_cols = ['gene', 'id', 'is_experimental', 'reference_seq_top_coverage', 'structure_file']

        for g in self.genes_with_a_representative_structure:
            repdict = g.protein.representative_structure.get_dict(df_format=True, only_keys=df_cols)
            repdict['gene'] = g.id
            rep_struct_pre_df.append(repdict)

        df = pd.DataFrame.from_records(rep_struct_pre_df, columns=df_cols).set_index('gene')
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df)

    @property
    def missing_representative_structure(self):
        return [x.id for x in self.genes if not self.genes_with_a_representative_structure.has_id(x.id)]

    def prep_itasser_models(self, itasser_installation, itlib_folder, runtype, create_in_dir=None,
                            execute_from_dir=None, all_genes=False, print_exec=False, **kwargs):
        """Prepare to run I-TASSER homology modeling for genes without structures, or all genes.

        Args:
            itasser_installation (str): Path to I-TASSER folder, i.e. '~/software/I-TASSER4.4'
            itlib_folder (str): Path to ITLIB folder, i.e. '~/software/ITLIB'
            runtype: How you will be running I-TASSER - local, slurm, or torque
            create_in_dir (str): Local directory where folders will be created
            execute_from_dir (str): Optional path to execution directory - use this if you are copying the homology
                models to another location such as a supercomputer for running
            all_genes (bool): If all genes should be prepped, or only those without any mapped structures
            print_exec (bool): If the execution statement should be printed to run modelling
            **kwargs:

        Todo:
            * kwargs - extra options for SLURM or Torque execution

        """
        # TODO: kwargs for slurm/torque options

        if not create_in_dir:
            if not self.data_dir:
                raise ValueError('Output directory must be specified')
            self.homology_models_dir = op.join(self.data_dir, 'homology_models')
        else:
            self.homology_models_dir = create_in_dir

        ssbio.utils.make_dir(self.homology_models_dir)

        if not execute_from_dir:
            execute_from_dir = self.homology_models_dir

        counter = 0
        for g in self.genes_with_a_representative_sequence:
            repseq = g.protein.representative_sequence

            # TODO: refactor to Protein class
            # Skip homology modeling if there is a structure set
            repstruct = g.protein.representative_structure
            if repstruct and not all_genes:
                log.debug('{}: representative structure set, skipping homology modeling'.format(g.id))
                continue
            if 'walltime' in kwargs:
                ITASSERPrep(ident=g.id, seq_str=repseq.seq_str, root_dir=self.homology_models_dir,
                            itasser_path=itasser_installation, itlib_path=itlib_folder,
                            runtype=runtype, print_exec=print_exec, execute_dir=execute_from_dir,
                            walltime=kwargs['walltime'])
            else:
                ITASSERPrep(ident=g.id, seq_str=repseq.seq_str, root_dir=self.homology_models_dir,
                            itasser_path=itasser_installation, itlib_path=itlib_folder,
                            runtype=runtype, print_exec=print_exec, execute_dir=execute_from_dir)
            counter += 1

        log.info('Prepared I-TASSER modeling folders for {} genes in folder {}'.format(counter,
                                                                                       self.homology_models_dir))

    def pdb_downloader_and_metadata(self, outdir=None, pdb_file_type=None, force_rerun=False):
        """Download all structures which have been mapped to our genes.

        Obtains the PDB file and mmCIF header and creates a metadata table "df_pdb_metadata".

        Args:
            outdir (str): Path to output directory of downloaded files, must be set if GEM-PRO directories
                were not created initially
            pdb_file_type (str): pdb, pdb.gz, mmcif, cif, cif.gz, xml.gz, mmtf, mmtf.gz - PDB structure file type that
                should be downloaded
            force_rerun (bool): If you want to overwrite any existing mappings and files

        """

        if not pdb_file_type:
            pdb_file_type = self.pdb_file_type

        counter = 0
        for g in tqdm(self.genes):
            pdbs = g.protein.pdb_downloader_and_metadata(outdir=outdir, pdb_file_type=pdb_file_type, force_rerun=force_rerun)

            if pdbs:
                counter += len(pdbs)

        log.info('Updated PDB metadata dataframe. See the "df_pdb_metadata" attribute for a summary dataframe.')
        log.info('Saved {} structures total'.format(counter))

    @property
    def df_pdb_metadata(self):
        df = pd.DataFrame()
        for g in self.genes_with_experimental_structures:
            # Get per protein DataFrame
            protein_df = g.protein.df_pdb_metadata.copy().reset_index()
            protein_df['gene'] = g.id
            df = df.append(protein_df)
        if df.empty:
            log.warning('Empty dataframe')
            return df
        else:
            return ssbio.utils.clean_df(df.set_index('gene'))

    def get_dssp_annotations(self):
        """Run DSSP on all representative structures and store calculations.

        Stored in::

            a_gene.protein.representative_structure.representative_chain.seq_record.letter_annotations['*-dssp']

        Todo:
            * Some errors arise from storing annotations for nonstandard amino acids, need to run DSSP separately for those

        """
        for g in tqdm(self.genes):
            if g.protein.representative_structure:
                try:
                    g.protein.representative_structure.get_dssp_annotations(outdir=g.protein.structure_dir)
                except PDBException as e:
                    print(e)
                    log.error('{}: Biopython error, issue matching sequences with {}'.format(g.id, g.protein.representative_structure))
                except TypeError as e:
                    print(e)
                    log.error('{}: Biopython error, DSSP SeqRecord length mismatch with {}'.format(g.id, g.protein.representative_structure))
                except Exception as e:
                    print(e)
                    log.error('{}: DSSP failed to run on {}'.format(g.id, g.protein.representative_structure))

    def get_msms_annotations(self):
        """Run MSMS on all representative structures and store calculations.

        Stored in::

            a_gene.protein.representative_structure.representative_chain.seq_record.letter_annotations['*-msms']

        """
        for g in tqdm(self.genes):
            if g.protein.representative_structure:
                try:
                    g.protein.representative_structure.get_residue_depths(outdir=g.protein.structure_dir)
                except TypeError:
                    log.error('{}: MSMS SeqRecord length mismatch with {}'.format(g.id, g.protein.representative_structure))
                except:
                    log.error(
                        '{}: unknown MSMS error with {}'.format(g.id, g.protein.representative_structure))

    def get_disulfide_bridges(self):
        """Run Biopython's disulfide bridge finder and store calculations.

        Stored in::

            a_gene.protein.representative_structure.representative_chain.seq_record.annotations['SSBOND-biopython']

        """
        for g in tqdm(self.genes):
            if g.protein.representative_structure:
                try:
                    g.protein.representative_structure.get_disulfide_bridges()
                except KeyError:
                    log.error('{}: unable to run disulfide bridge finder on {}'.format(g.id, g.protein.representative_structure))

    ### END STRUCTURE RELATED METHODS ###
    ####################################################################################################################

    # def run_pipeline(self, sequence_mapping_engine='', structure_mapping_engine='', **kwargs):
    #     """Run the entire GEM-PRO pipeline.
    #
    #     Options include:
    #     ...
    #
    #     Returns:
    #
    #     """
    #     current_sequence_mapping_engines = ['kegg', 'uniprot', 'all']
    #     if sequence_mapping_engine not in current_sequence_mapping_engines:
    #         raise ValueError('Sequence mapping engine not available')
    #
    #     if sequence_mapping_engine == 'kegg' or sequence_mapping_engine == 'all':
    #         if 'kegg_organism_code' not in kwargs:
    #             raise TypeError('kegg_organism_code needed')
    #
    #     if sequence_mapping_engine == 'uniprot' or sequence_mapping_engine == 'all':
    #         if 'model_gene_source' not in kwargs:
    #             raise TypeError('UniProt model_gene_source needed')
    #
    #     print('Running KEGG mapping...')
    #     # TODO: better way of passing optional arguments to these things - yaml? There are too many if statements to check for things
    #     if 'custom_gene_mapping' in kwargs:
    #         self.kegg_mapping_and_metadata(kegg_organism_code=kwargs['kegg_organism_code'], custom_gene_mapping=kwargs['custom_gene_mapping'])
    #     else:
    #         self.kegg_mapping_and_metadata(kegg_organism_code=kwargs['kegg_organism_code'])
    #
    #     print('Running UniProt mapping...')
    #     self.uniprot_mapping_and_metadata(model_gene_source=kwargs['model_gene_source'])
    #
    #     print('Setting representative sequences...')
    #     self.set_representative_sequence()
    #
    #     current_structure_mapping_engines = ['uniprot', 'itasser', 'all', 'none'] # 'blast', 'manual_homology'
    #     if structure_mapping_engine not in current_structure_mapping_engines:
    #         raise ValueError('Structure mapping engine not available')
    #
    #     if structure_mapping_engine == 'itasser' or structure_mapping_engine == 'all':
    #         if 'homology_raw_dir' not in kwargs:
    #             raise TypeError('Path to homology models needed as homology_raw_dir')
    #
    #     if not structure_mapping_engine == 'none':
    #         if 'always_use_homology' not in kwargs:
    #             raise TypeError('Flag always_use_homology needed')
    #
    #     if structure_mapping_engine == 'uniprot' or structure_mapping_engine == 'all':
    #         print('Mapping UniProt IDs to the PDB...')
    #         self.map_uniprot_to_pdb()
    #
    #     if structure_mapping_engine == 'itasser' or structure_mapping_engine == 'all':
    #         print('Copying I-TASSER models...')
    #         self.get_itasser_models(homology_raw_dir=kwargs['homology_raw_dir'],
    #                                 custom_itasser_name_mapping=kwargs['custom_itasser_name_mapping'])
    #
    #     if not structure_mapping_engine == 'none':
    #         print('Setting representative structures...')
    #         self.set_representative_structure(always_use_homology=kwargs['always_use_homology'])

    def __json_encode__(self):
        to_return = {}
        # Don't save properties, methods in the JSON
        for x in [a for a in dir(self) if not a.startswith('__') and not isinstance(getattr(type(self), a, None), property) and not callable(getattr(self,a))]:
            if self.model and x == 'genes':
                continue
            to_return.update({x: getattr(self, x)})
        return to_return

    def __json_decode__(self, **attrs):
        for k, v in attrs.items():
            setattr(self, k, v)
        if not self.model:
            self.genes = DictList(self.genes)
        else:
            self.genes = self.model.genes


if __name__ == '__main__':
    pass
    # TODO: Run the GEM-PRO pipeline!


#!/usr/bin/env python
import argparse
import glob
import os
import os.path as op
import pandas as pd
from tqdm import tqdm
from ssbio import utils
date = utils.Date()

import math
# import cachetools
import scipy.constants

# author: Ke Chen

# R = scipy.constants.R # J/K.mol
# R is the molar gas constant in cal/K.mol
r_cal = scipy.constants.R / scipy.constants.calorie

# Oobatake dG
# dG and dCp from Table 8 in Oobatake paper.
# dG,dH in unit kcal/mol
# dCp,dS in unit cal/mol.K
oobatake_dictionary = {}
oobatake_dictionary['A'] = {'dG': -0.02, 'dCp': 14.22, 'dH': 0.51, 'dS': 1.82}
oobatake_dictionary['C'] = {'dG': 1.08, 'dCp': 9.41, 'dH': 5.21, 'dS': 13.85}
oobatake_dictionary['D'] = {'dG': -0.08, 'dCp': 2.73, 'dH': 0.18, 'dS': 0.86}
oobatake_dictionary['E'] = {'dG': -0.13, 'dCp': 3.17, 'dH': 0.05, 'dS': 0.65}
oobatake_dictionary['F'] = {'dG': 2.16, 'dCp': 39.06, 'dH': 6.82, 'dS': 15.64}
oobatake_dictionary['G'] = {'dG': 0.09, 'dCp': 4.88, 'dH': -0.23, 'dS': -1.07}
oobatake_dictionary['H'] = {'dG': 0.56, 'dCp': 20.05, 'dH': 0.79, 'dS': 0.75}
oobatake_dictionary['I'] = {'dG': -0.08, 'dCp': 41.98, 'dH': 0.19, 'dS': 0.89}
oobatake_dictionary['K'] = {'dG': -0.32, 'dCp': 17.68, 'dH': -1.45, 'dS': -3.78}
oobatake_dictionary['L'] = {'dG': -0.08, 'dCp': 38.26, 'dH': 0.17, 'dS': 0.81}
oobatake_dictionary['M'] = {'dG': 0.53, 'dCp': 31.67, 'dH': 2.89, 'dS': 7.93}
oobatake_dictionary['N'] = {'dG': -0.30, 'dCp': 3.91, 'dH': -2.03, 'dS': -5.83}
oobatake_dictionary['P'] = {'dG': -0.06, 'dCp': 23.69, 'dH': 0.02, 'dS': 0.30}
oobatake_dictionary['Q'] = {'dG': -0.23, 'dCp': 3.74, 'dH': -1.76, 'dS': -5.13}
oobatake_dictionary['R'] = {'dG': -0.71, 'dCp': 16.66, 'dH': -4.40, 'dS': -12.38}
oobatake_dictionary['S'] = {'dG': -0.40, 'dCp': 6.14, 'dH': -0.16, 'dS': 0.84}
oobatake_dictionary['T'] = {'dG': -0.24, 'dCp': 16.11, 'dH': 0.04, 'dS': 0.89}
oobatake_dictionary['V'] = {'dG': -0.06, 'dCp': 32.58, 'dH': 0.30, 'dS': 1.20}
oobatake_dictionary['W'] = {'dG': 1.78, 'dCp': 37.69, 'dH': 4.47, 'dS': 9.00}
oobatake_dictionary['Y'] = {'dG': 0.91, 'dCp': 30.54, 'dH': 3.73, 'dS': 9.46}
oobatake_dictionary['U'] = {'dG': 1.08, 'dCp': 9.41, 'dH': 5.21, 'dS': 13.85}  # assume U==C


def sum_of_dCp(seq):
    dCp_sum = 0
    for aa in seq:
        dCp_sum += oobatake_dictionary[aa]['dCp']
    return dCp_sum


def calculate_oobatake_dH(seq, temp):
    """Get dH in units cal/mol

    Args:
        seq: amino acid sequence
        temp: temperature in degrees C

    Returns:
        dH

    """
    dH = 0
    temp += 273.15
    T0 = 298.15
    for aa in seq:
        H0 = oobatake_dictionary[aa]['dH'] * 1000
        dH += H0
    return dH + sum_of_dCp(seq) * (temp - T0)


def calculate_oobatake_dS(seq, temp):
    """Get dS in units cal/mol

    Args:
        seq: amino acid sequence
        temp: temperature in degrees C

    Returns:
        dS

    """
    dS = 0
    temp += 273.15
    T0 = 298.15
    dCp_sum = sum_of_dCp(seq)
    for aa in seq:
        S0 = oobatake_dictionary[aa]['dS']
        dS += S0
    return dS + dCp_sum * math.log(temp / T0)


# @cachetools.func.ttl_cache(maxsize=256)
def calculate_oobatake_dG(seq, temp):
    """Get dG in units cal/mol

    Args:
        seq: amino acid sequence
        temp: temperature in degrees C

    Returns:
        dG

    """
    T0 = 298.15
    dH = calculate_oobatake_dH(seq, temp)
    dS = calculate_oobatake_dS(seq, temp)
    dG = dH - (temp + 273.15) * dS

    # 563.552 - a correction for N- and C-terminal group (approximated from 7 examples in the paper)
    return dG - 563.552


# @cachetools.func.ttl_cache(maxsize=128)
def calculate_dill_dG(seq_len, temp):
    """Get dG using Dill method in units J/mol

    Args:
        seq_len: length of amino acid sequence
        temp: temperature in degrees C

    Returns:
        dG in J/mol
    """
    Th = 373.5  # this quantity affects the up-and-down of the dG vs temperature curve (dG values)
    Ts = 385  # this quantity affects the left-and-right
    temp += 273.15
    dH = (4.0 * seq_len + 143) * 1000
    dS = 13.27 * seq_len + 448
    dCp = (0.049 * seq_len + 0.85) * 1000
    dG = dH + dCp * (temp - Th) - temp * dS - temp * dCp * math.log(float(temp) / Ts)
    return dG

# @cachetools.func.ttl_cache(maxsize=500)
def get_dG_at_T(seq, temp):
    """Predict dG at temperature T

    Args:
        seq: sequence string
        temp: temperature

    Returns:
        free energy of unfolding dG (cal/mol)
        equilibrium constant Keq (for ME2.0 coupling constraint)
        method used to calculate
    """

    oobatake = {}
    for t in range(20, 51):
        oobatake[t] = calculate_oobatake_dG(seq, t)

    stable = [i for i in oobatake.values() if i > 0]
    if len(stable) == 0:
        # if oobatake dG < 0 for all tempertures [20,50], use Dill dG
        # and convert the number from J/mol to cal/mol
        dG = 0.238846 * calculate_dill_dG(len(seq), temp)
        method='Dill'
    else:
        dG = oobatake[temp]
        method='Oobatake'

    keq = math.exp(-1 * dG / (r_cal * (temp + 273.15)))

    return dG, keq, method


if __name__ == '__main__':
    # load inputs from command line

    p = argparse.ArgumentParser(description='Run thermostability calculations on a FASTA file or a folder of FASTA files.')
    p.add_argument('infile', help='FASTA file or directory of FASTA files.')
    p.add_argument('--temp', '-t', default=37)
    args = p.parse_args()


    prop_dir = 'properties'
    if not op.exists(prop_dir):
        os.mkdir(prop_dir)

    # TODO: improve arg parsing for files/dirs
    if len(args.infile) == 1 and op.isdir(args.infile[0]):
        os.chdir(args.infile[0])
        files = glob.glob('*')
    else:
        files = args.infile

    results = []

    for file in tqdm(files):
        if op.isdir(file):
            continue

        # load the sequence file, also the ID
        seq_records = fasta.load_fasta_file(file)
        seq_id = op.splitext(op.basename(file))[0]

        # TODO: seems useless to return seqrecords to just convert them to strings
        for seq_record in seq_records:
            res = ts.get_dG_at_T(str(seq_record.seq), temp=args.temp)
            result = {'id':seq_id, 'dg':res[0], 'keq':res[1], 'method':res[2]}
            results.append(result)

    agg_df = pd.DataFrame(results)
    agg_df.to_csv(op.join(prop_dir, '{}_thermo_results.csv'.format(date.short_date)))
    print('Saved results in thermo_results.csv')


from collections import defaultdict
import logging
log = logging.getLogger(__name__)

# Parsing TMHMM results
# There are two output formats: Long and short. Long output format
#
# For the long format (default), tmhmm gives some statistics and a list of the location of the predicted transmembrane helices and the predicted location of the intervening loop regions. Here is an example:
#
#     # COX2_BACSU Length: 278
#     # COX2_BACSU Number of predicted TMHs: 3
#     # COX2_BACSU Exp number of AAs in TMHs: 68.6888999999999
#     # COX2_BACSU Exp number, first 60 AAs: 39.8875
#     # COX2_BACSU Total prob of N-in: 0.99950
#     # COX2_BACSU POSSIBLE N-term signal sequence
#     COX2_BACSU TMHMM2.0 inside 1 6
#     COX2_BACSU TMHMM2.0 TMhelix 7 29
#     COX2_BACSU TMHMM2.0 outside 30 43
#     COX2_BACSU TMHMM2.0 TMhelix 44 66
#     COX2_BACSU TMHMM2.0 inside 67 86
#     COX2_BACSU TMHMM2.0 TMhelix 87 109
#     COX2_BACSU TMHMM2.0 outside 110 278
#
# If the whole sequence is labeled as inside or outside, the prediction is that it contains no membrane
# helices. It is probably not wise to interpret it as a prediction of location. The prediction gives the most probable location and orientation of transmembrane helices in the sequence. It is found by an algorithm called N-best (or 1-best in this case) that sums over all paths through the model with the same location and direction of the helices.
#
# The first few lines gives some statistics:
#
# Length: the length of the protein sequence.
#
# Number of predicted TMHs: The number of predicted transmembrane helices.
#
# Exp number of AAs in TMHs: The expected number of amino acids intransmembrane helices. If this number is larger than 18 it is very likely to be a transmembrane protein (OR have a signal peptide).
#
# Exp number, first 60 AAs: The expected number of amino acids in transmembrane helices in the first 60 amino acids of the protein. If this number more than a few, you should be warned that a predicted transmembrane helix in the N-term could be a signal peptide.
#
# Total prob of N-in: The total probability that the N-term is on the cytoplasmic side of the membrane.
#
# POSSIBLE N-term signal sequence: a warning that is produced when "Exp number, first 60 AAs" is larger than 10.


def parse_tmhmm_long(tmhmm_results):
    with open(tmhmm_results) as f:
        lines = f.read().splitlines()

    infodict = defaultdict(dict)

    for l in lines:
        if 'Number of predicted TMHs:' in l:
            gene = l.split(' Number')[0].strip('# ')
            infodict[gene]['num_tm_helices'] = int(l.split(': ')[1])

        if 'WARNING' in l:
            log.warning('{}: no TMHMM predictions'.format(l))
            continue

        # TODO: POSSIBLE N-term signal sequence - parse this

        # Look for the lines without #, these are the TM predicted regions
        if '#' not in l:
            stuff = l.split()
            if stuff[1] == 'TMHMM2.0':

                gene = stuff[0]
                region = stuff[2]
                region_start = stuff[3]
                region_end = stuff[4]

                if 'sequence' in infodict[gene]:
                    tm_seq = infodict[gene]['sequence']
                else:
                    tm_seq = ''

                if region == 'outside':
                    info = 'O'
                elif region == 'inside':
                    info = 'I'
                elif region == 'TMhelix':
                    info = 'T'
                else:
                    log.error('{}: unknown region type'.format(info))
                    info = '-'

                for r in range(int(region_start), int(region_end) + 1):
                    tm_seq += info

                infodict[gene]['sequence'] = tm_seq

    return infodict


def label_TM_tmhmm_residue_numbers_and_leaflets(tmhmm_seq):
    """Determine the residue numbers of the TM-helix residues that cross the membrane and label them by leaflet.

    Args:
        tmhmm_seq: g.protein.representative_sequence.seq_record.letter_annotations['TM-tmhmm']

    Returns:
        leaflet_dict: a dictionary with leaflet_variable : [residue list] where the variable is inside or outside
        TM_boundary dict: outputs a dictionar with : TM helix number : [TM helix residue start , TM helix residue end]

    TODO:
        untested method!
    """

    TM_number_dict = {}
    T_index = []
    T_residue = []

    residue_count = 1
    for residue_label in tmhmm_seq:
        if residue_label == 'T':
            T_residue.append(residue_count)

        residue_count = residue_count + 1
    TM_number_dict.update({'T_residue': T_residue})

    # finding the TM boundaries
    T_residue_list = TM_number_dict['T_residue']

    count = 0
    max_count = len(T_residue_list) - 1
    TM_helix_count = 0
    TM_boundary_dict = {}

    while count <= max_count:
        # first residue = TM start
        if count == 0:
            TM_start = T_residue_list[count]
            count = count + 1
            continue
        # Last residue = TM end
        elif count == max_count:
            TM_end = T_residue_list[count]
            TM_helix_count = TM_helix_count + 1
            TM_boundary_dict.update({'TM_helix_' + str(TM_helix_count): [TM_start, TM_end]})
            break
        # middle residues need to be start or end
        elif T_residue_list[count] != T_residue_list[count + 1] - 1:
            TM_end = T_residue_list[count]
            TM_helix_count = TM_helix_count + 1
            TM_boundary_dict.update({'TM_helix_' + str(TM_helix_count): [TM_start, TM_end]})
            # new TM_start
            TM_start = T_residue_list[count + 1]
        count = count + 1
    # assign leaflet to proper TM residues O or I
    leaflet_dict = {}
    for leaflet in ['O', 'I']:
        leaflet_list = []
        for TM_helix, TM_residues in TM_boundary_dict.iteritems():
            for residue_num in TM_residues:
                tmhmm_seq_index = residue_num - 1
                previous_residue = tmhmm_seq_index - 1
                next_residue = tmhmm_seq_index + 1
                # identify if the previous or next residue closest to the TM helix start/end is the proper leaflet
                if tmhmm_seq[previous_residue] == leaflet or tmhmm_seq[next_residue] == leaflet:
                    leaflet_list.append(residue_num)
        leaflet_dict.update({'tmhmm_leaflet_' + leaflet: leaflet_list})

    return TM_boundary_dict, leaflet_dict


import filecmp
import os
import os.path as op
import unittest

import ssbio.protein.sequence.utils.fasta as fasta


class TestFasta(unittest.TestCase):
    """Unit tests for sequence/fasta tools."""

    def test_load_fasta_file(self):
        sequences = fasta.load_fasta_file(op.join('test_files', 'sequences', 'X5D299-1.faa'))
        self.assertEqual(len(sequences), 1)
        self.assertEqual(sequences[0].id, 'X5D299-1')
        self.assertRaises(IOError, fasta.load_fasta_file, 'NOTAFILE.faa')

    def test_load_seq_string_as_seqrecord(self):
        test_string = 'ASQAGIPSGVYNVIPCSRKNAKEVGEAICTDPLVSKISF'
        ident = 'TEST'
        sequence = fasta.load_seq_string_as_seqrecord(test_string, ident)
        self.assertEqual(len(sequence), len(test_string))
        self.assertEqual(sequence.id, ident)

        self.assertRaises(ValueError, fasta.load_seq_string_as_seqrecord, 'N-0t,V4L1DAA', ident)

    def test_write_fasta_file(self):
        test_string = 'ASQAGIPSGVYNVIPCSRKNAKEVGEAICTDPLVSKISF'
        ident = 'TEST'
        sequence = fasta.load_seq_string_as_seqrecord(test_string, ident, desc='tester')
        custom_ext = '.fasta'
        custom_outname = 'TESTER'
        overwrite = True

        custom_dir = op.join('test_files', 'out')
        gold_standard = op.join('test_files', 'sequences', 'TESTER.faa')

        written_file_cusdir = fasta.write_fasta_file(sequence, custom_outname,
                                                     outdir=custom_dir, force_rerun=overwrite)
        self.assertEqual(written_file_cusdir, os.path.join(custom_dir, custom_outname + '.faa'))
        # Test is file contents are equal with filecmp
        self.assertTrue(filecmp.cmp(gold_standard, written_file_cusdir))

        written_file_cusdirid = fasta.write_fasta_file(sequence, custom_outname,
                                                       outdir=custom_dir, force_rerun=overwrite)
        self.assertEqual(written_file_cusdirid, os.path.join(custom_dir, custom_outname + '.faa'))
        self.assertTrue(filecmp.cmp(gold_standard, written_file_cusdirid))

        written_file_cusdiridext = fasta.write_fasta_file(sequence, custom_outname,
                                                          outdir=custom_dir, outext=custom_ext, force_rerun=overwrite)
        self.assertEqual(written_file_cusdiridext, os.path.join(custom_dir, custom_outname + custom_ext))
        self.assertTrue(filecmp.cmp(gold_standard, written_file_cusdiridext))

    def test_write_fasta_file_from_dict(self):
        tester = {'TEST': 'ASQAGIPSGVYNVIPCSRKNAKEVGEAICTDPLVSKISF'}
        custom_ext = '.fasta'
        custom_outname = 'TESTER'
        outdir = op.join('test_files', 'out')

        written_file_cusdir = fasta.write_fasta_file_from_dict(tester, custom_outname,
                                                               outdir=outdir, force_rerun=True)
        self.assertEqual(written_file_cusdir, os.path.join(outdir, custom_outname + '.faa'))

        written_file_cusdiridext = fasta.write_fasta_file_from_dict(tester, custom_outname,
                                                                    outdir=outdir, outext=custom_ext, force_rerun=True)
        self.assertEqual(written_file_cusdiridext, os.path.join(outdir, custom_outname + custom_ext))

if __name__ == "__main__":
    unittest.main()



import logging
import os.path as op
from Bio import Entrez
from ssbio import utils

date = utils.Date()
log = logging.getLogger(__name__)


def download_genome_sequence(genome_accession_or_id, seqtype, email, outdir='', outfile='', force_rerun=False):
    """Download the entire set of DNA or protein sequences from protein-encoding genes in a genome from NCBI.

    Saves a FASTA file in the optional directory specified.

    Args:
        genome_accession_or_id (str): RefSeq complete genome ID (e.g. "NC_000913") or GenBank ID (e.g. "U00096.3")
        seqtype (str): "dna" or "protein" - if you want the coding sequences in amino acid or DNA format.
        email (str): mandatory email so NCBI knows who is accessing the execute_dir
        outdir (str): optional output directory (default is the current directory)

    Returns:
        Path to downloaded FASTA file.

    """
    if seqtype == 'dna':
        extension = 'fna'
        rettype = "fasta_cds_na"
    elif seqtype == 'protein':
        extension = 'faa'
        rettype = "fasta_cds_aa"
    else:
        raise ValueError('seqtype must be "dna" or "protein"')

    # path and filename parsing
    if outfile:
        outfile = op.join(outdir, '{}.{}'.format(outfile, extension))
    else:
        # if no outfile is specified, default is "$GI.faa"
        outfile = op.join(outdir, '{}.{}'.format(genome_accession_or_id, extension))

    if not force_rerun:
        # return the path to the file if it was already downloaded
        if op.exists(outfile):
            log.debug('FASTA file already exists at {}'.format(outfile))
            return outfile

    Entrez.email = email

    # convert to NCBI GI
    convert_to_gi = Entrez.read(Entrez.esearch(db="nucleotide", term=genome_accession_or_id, retmode="xml"))
    gi = convert_to_gi['IdList']

    # download only the CDS protein FASTA
    # See: https://www.ncbi.nlm.nih.gov/books/NBK25499/table/chapter4.T._valid_values_of__retmode_and/
    record = Entrez.efetch(db="nucleotide", id=gi, rettype=rettype, retmode="fasta")

    # save the FASTA file
    with open(outfile, 'wt') as f:
        f.write(record.read())

    log.info('Saved FASTA file at {}'.format(outfile))
    return outfile


import os.path as op
import unittest

from ssbio.protein.sequence.seqprop import SeqProp


class TestSeqProp(unittest.TestCase):
    """Unit tests for SeqProp"""

    @classmethod
    def setUpClass(self):
        self.sp = SeqProp(ident='P0ABP8', sequence_path='test_files/sequences/P0ABP8.fasta',
                          metadata_path='test_files/sequences/P0ABP8.txt')

    def test_load_seq_file(self):
        self.assertEqual(self.sp.sequence_file, 'P0ABP8.fasta')
        self.assertEqual(self.sp.seq_len, 239)

    def test_load_metadata_file(self):
        self.assertEqual(self.sp.metadata_file, 'P0ABP8.txt')

    def test_get_seq_str(self):
        seq = 'MATPHINAEMGDFADVVLMPGDPLRAKYIAETFLEDAREVNNVRGMLGFTGTYKGRKISVMGHGMGIPSCSIYTKELITDFGVKKIIRVGSCGAVLPHVKLRDVVIGMGACTDSKVNRIRFKDHDFAAIADFDMVRNAVDAAKALGIDARVGNLFSADLFYSPDGEMFDVMEKYGILGVEMEAAGIYGVAAEFGAKALTICTVSDHIRTHEQTTAAERQTTFNDMIKIALESVLLGDKE'
        self.assertTrue(self.sp.seq_str)
        self.assertEqual(self.sp.seq_str, seq)

    def test_equal_to(self):
        newsp = SeqProp(ident='P0ABP8', sequence_path='test_files/sequences/P0ABP8.fasta',
                        metadata_path='test_files/sequences/P0ABP8.txt')
        self.assertTrue(self.sp.equal_to(newsp))

    def test_equal_to_fasta(self):
        self.assertTrue(self.sp.equal_to_fasta('test_files/sequences/P0ABP8.fasta'))

    def test_num_pdbs(self):
        self.assertEqual(self.sp.num_pdbs, 0)
        self.sp.pdbs = ['1abc', '2abc']
        self.assertEqual(self.sp.num_pdbs, 2)

    def test_write_fasta_file(self):
        self.sp.write_fasta_file(outfile='test_files/out/P0ABP8_new.faa', force_rerun=True)
        self.assertTrue(op.exists('test_files/out/P0ABP8_new.faa'))
        self.assertEqual(self.sp.sequence_file, 'P0ABP8_new.faa')


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
        name='ssbio',
        version='0.9.2',
        author='Nathan Mih',
        author_email='nmih@ucsd.edu',
        license='MIT',
        url='http://github.com/SBRG/ssbio',
        download_url = 'https://github.com/SBRG/ssbio/archive/v0.9.2.tar.gz',
        description='Tools to enable structural systems biology',
        packages=find_packages(),
        package_dir={'ssbio': 'ssbio'},
        package_data={'ssbio': ['ssbio/etc/*']},
        scripts=['ssbio/protein/structure/utils/cleanpdb.py',
                 'ssbio/protein/sequence/properties/aggregation_propensity.py',
                 'ssbio/protein/sequence/properties/thermostability.py',
                 'ssbio/protein/structure/utils/mutatepdb.py',
                 'ssbio/protein/structure/utils/tleap.py',
                 'ssbio/databases/drugbank.py',
                 'ssbio/protein/structure/properties/msms.py'],
        long_description=open('README.md').read(),
        install_requires=required
)



import requests


def chembl_uniprot_to_chembl_id(uniprot_id):
    """Map a UniProt ID to a ChEMBL ID.

    Args:
        uniprot_id (str): UniProt ID

    Returns:
        str: ChEMBL ID

    """
    r = requests.get('https://www.ebi.ac.uk/chemblws/targets/uniprot/{}.json'.format(uniprot_id))
    if r.status_code == 200:
        d = r.json()
        return str(d['target']['chemblId'])
    else:
        return None


def chembl_targets_to_approved_drugs(chembl_id):
    """Map a ChEMBL ID to the ChEMBL drugs that target it

    Args:
        chembl_id (str):

    Returns:
        ?

    """
    r = requests.get('https://www.ebi.ac.uk/chemblws/targets/{}/approvedDrug.json'.format(chembl_id))
    if r.status_code == 200:
        d = r.json()
        approved_drugs = d['approvedDrugs']
        if approved_drugs:
            drugs_names_actions = {}
            for drug in approved_drugs:
                drug_id = str(drug['chemblId'])
                drug_name = str(drug['name'])
                drug_mech = str(drug['mechanismOfAction'])
                drugs_names_actions[(drug_id,drug_name)] = drug_mech
            return drugs_names_actions
        else:
            return None


def chembl_uniprot_to_drugs(uniprot_id):
    """Map a UniProt ID to approved drugs that target it

    Args:
        uniprot_id (str): UniProt ID

    Returns:
        ?

    """
    chembl_id = chembl_uniprot_to_chembl_id(uniprot_id)
    if chembl_id:
        return chembl_targets_to_approved_drugs(chembl_id)
    else:
        return None


import logging
from collections import defaultdict

import pandas as pd
from Bio import PDB
from Bio.PDB.DSSP import dssp_dict_from_pdb_file
from Bio.PDB.DSSP import residue_max_acc
from Bio.PDB.Polypeptide import aa1
from Bio.PDB.Polypeptide import one_to_three

import ssbio.utils

log = logging.getLogger(__name__)


def get_dssp_df(model, pdb_file, outfile=None, outdir=None, outext='_dssp.df', force_rerun=False):
    """

    Args:
        model:
        pdb_file:
        outfile:
        outdir:
        outext:
        force_rerun:

    Returns:

    """
    # Create the output file name
    outfile = ssbio.utils.outfile_maker(inname=pdb_file, outname=outfile, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        try:
            # TODO: errors with non-standard residues, ie. MSE in 4Q6U or 1nfr
            # TODO: write command line executor for DSSP and parser for raw DSSP results
            dssp = PDB.DSSP(model, pdb_file)
        except KeyError:
            return pd.DataFrame()

        if len(dssp.property_list) == 0:
            return pd.DataFrame()

        # Reorganize the results into a csv file
        appender = []
        for k in dssp.property_keys:
            to_append = []
            x = dssp.property_dict[k]
            chain = k[0]
            residue = k[1]
            het = residue[0]
            resnum = residue[1]
            icode = residue[2]
            to_append.extend([chain, resnum, icode])
            to_append.extend(x)
            appender.append(to_append)

        cols = ['chain', 'resnum', 'icode',
                'dssp_index', 'aa', 'ss', 'exposure_rsa', 'phi', 'psi',
                'NH_O_1_relidx', 'NH_O_1_energy', 'O_NH_1_relidx',
                'O_NH_1_energy', 'NH_O_2_relidx', 'NH_O_2_energy',
                'O_NH_2_relidx', 'O_NH_2_energy']

        df = pd.DataFrame.from_records(appender, columns=cols)

        # Adding additional columns
        df = df[df['aa'].isin(list(aa1))]
        df['aa_three'] = df['aa'].apply(one_to_three)
        df['max_acc'] = df['aa_three'].map(residue_max_acc['Sander'].get)
        df[['exposure_rsa', 'max_acc']] = df[['exposure_rsa', 'max_acc']].astype(float)
        df['exposure_asa'] = df['exposure_rsa'] * df['max_acc']

        df.to_csv(outfile)
    else:
        log.debug('{}: already ran DSSP and force_rerun={}, loading results'.format(outfile, force_rerun))
        df = pd.read_csv(outfile, index_col=0)

    return df


def get_dssp_df_on_file(pdb_file, outfile=None, outdir=None, outext='_dssp.df', force_rerun=False):
    """Run DSSP directly on a structure file with the Biopython method Bio.PDB.DSSP.dssp_dict_from_pdb_file

    Avoids errors like: PDBException: Structure/DSSP mismatch at <Residue MSE het=  resseq=19 icode= >
        by not matching information to the structure file (DSSP fills in the ID "X" for unknown residues)

    Args:
        pdb_file: Path to PDB file
        outfile: Name of output file
        outdir: Path to output directory
        outext: Extension of output file
        force_rerun: If DSSP should be rerun if the outfile exists

    Returns:
        Pandas DataFrame: DSSP results, summarized

    """
    # TODO: function unfinished
    # Create the output file name
    outfile = ssbio.utils.outfile_maker(inname=pdb_file, outname=outfile, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        try:
            d = dssp_dict_from_pdb_file(pdb_file)
        except Exception('DSSP failed to produce an output'):
            log.error('{}: unable to run DSSP'.format(pdb_file))
            return pd.DataFrame()

        appender = []
        # TODO: WARNING: d is slightly different than when using function get_dssp_df
        for k in d[1]:
            to_append = []
            y = d[0][k]
            chain = k[0]
            residue = k[1]
            het = residue[0]
            resnum = residue[1]
            icode = residue[2]
            to_append.extend([chain, resnum, icode])
            to_append.extend(y)
            appender.append(to_append)

        cols = ['chain', 'resnum', 'icode',
                'dssp_index', 'aa', 'ss', 'exposure_rsa', 'phi', 'psi',
                'NH_O_1_relidx', 'NH_O_1_energy', 'O_NH_1_relidx',
                'O_NH_1_energy', 'NH_O_2_relidx', 'NH_O_2_energy',
                'O_NH_2_relidx', 'O_NH_2_energy']

        df = pd.DataFrame.from_records(appender, columns=cols)

        # Adding additional columns
        df = df[df['aa'].isin(list(aa1))]
        df['aa_three'] = df['aa'].apply(one_to_three)
        df['max_acc'] = df['aa_three'].map(residue_max_acc['Sander'].get)
        df[['exposure_rsa', 'max_acc']] = df[['exposure_rsa', 'max_acc']].astype(float)
        df['exposure_asa'] = df['exposure_rsa'] * df['max_acc']

        df.to_csv(outfile)
    else:
        log.debug('{}: already ran DSSP and force_rerun={}, loading results'.format(outfile, force_rerun))
        df = pd.read_csv(outfile, index_col=0)

    return df

def secondary_structure_summary(dssp_df):
    """Summarize the secondary structure content of the DSSP dataframe for each chain.

    Args:
        dssp_df: Pandas DataFrame of parsed DSSP results

    Returns:
        dict: Chain to secondary structure summary dictionary

    """
    chains = dssp_df.chain.unique()

    infodict = {}
    for chain in chains:
        expoinfo = defaultdict(int)
        chain_df = dssp_df[dssp_df.chain == chain]
        counts = chain_df.ss.value_counts()
        total = float(len(chain_df))

        for ss, count in counts.items():
            if ss == '-':
                expoinfo['percent_C-dssp'] = count/total
            if ss == 'H':
                expoinfo['percent_H-dssp'] = count/total
            if ss == 'B':
                expoinfo['percent_B-dssp'] = count/total
            if ss == 'E':
                expoinfo['percent_E-dssp'] = count/total
            if ss == 'G':
                expoinfo['percent_G-dssp'] = count/total
            if ss == 'I':
                expoinfo['percent_I-dssp'] = count/total
            if ss == 'T':
                expoinfo['percent_T-dssp'] = count/total
            if ss == 'S':
                expoinfo['percent_S-dssp'] = count/total

        # Filling in 0 percenters
        for per in ['percent_C-dssp','percent_H-dssp','percent_B-dssp','percent_E-dssp',
                    'percent_G-dssp','percent_I-dssp','percent_T-dssp','percent_S-dssp']:
            if per not in expoinfo:
                expoinfo[per] = 0.0

        infodict[chain] = dict(expoinfo)

    return infodict


# TODO: below methods have not been fixed
def calc_surface_buried(dssp_df):
    '''Calculates the percent of residues that are in the surface or buried,
    as well as if they are polar or nonpolar. Returns a dictionary of this.
    '''
    SN = 0
    BN = 0
    SP = 0
    SNP = 0
    SPo = 0
    SNe = 0
    BNP = 0
    BP = 0
    BPo = 0
    BNe = 0
    Total = 0

    sbinfo = {}

    df_min = dssp_df[['aa_three', 'exposure_asa']]

    if len(df_min) == 0:
        return sbinfo
    else:
        for i, r in df_min.iterrows():
            res = r.aa_three
            area = r.exposure_asa
            if res in AAdict:
                if AAdict[res] == 'nonpolar' and area > 3:
                    SNP = SNP + 1
                    SN = SN + 1
                elif AAdict[res] == 'polar' and area > 3:
                    SP = SP + 1
                    SN = SN + 1
                elif AAdict[res] == 'positive' and area > 3:
                    SPo = SPo + 1
                    SN = SN + 1
                elif AAdict[res] == 'negative' and area > 3:
                    SNe = SNe + 1
                    SN = SN + 1
                elif AAdict[res] == 'positive' and area <= 3:
                    BPo = BPo + 1
                    BN = BN + 1
                elif AAdict[res] == 'negative' and area <= 3:
                    BNe = BNe + 1
                    BN = BN + 1
                elif AAdict[res] == 'polar' and area <= 3:
                    BP = BP + 1
                    BN = BN + 1
                elif AAdict[res] == 'nonpolar' and area <= 3:
                    BNP = BNP + 1
                    BN = BN + 1
        Total = float(BN + SN)
        pSNP = float(SNP) / Total
        pSP = float(SP) / Total
        pSPo = float(SPo) / Total
        pSNe = float(SNe) / Total
        pBNP = float(BNP) / Total
        pBP = float(BP) / Total
        pBPo = float(BPo) / Total
        pBNe = float(BNe) / Total
        pBN = float(BN) / Total
        pSN = float(SN) / Total
        sbinfo['ssb_per_S_NP'] = pSNP
        sbinfo['ssb_per_S_P'] = pSP
        sbinfo['ssb_per_S_pos'] = pSPo
        sbinfo['ssb_per_S_neg'] = pSNe
        sbinfo['ssb_per_B_NP'] = pBNP
        sbinfo['ssb_per_B_P'] = pBP
        sbinfo['ssb_per_B_pos'] = pBPo
        sbinfo['ssb_per_B_neg'] = pBNe
        sbinfo['ssb_per_S'] = pSN
        sbinfo['ssb_per_B'] = pBN

        return sbinfo


def calc_sasa(dssp_df):
    """
    Calculation of SASA utilizing the DSSP program.

    DSSP must be installed for biopython to properly call it.
    Install using apt-get on Ubuntu
    or from: http://swift.cmbi.ru.nl/gv/dssp/

    Input: PDB or CIF structure file
    Output: SASA (integer) of structure
    """

    infodict = {'ssb_sasa': dssp_df.exposure_asa.sum(),
                'ssb_mean_rel_exposed': dssp_df.exposure_rsa.mean(),
                'ssb_size': len(dssp_df)}

    return infodict


def all_dssp_props(filename, file_type):
    '''Returns a large dictionary of SASA, secondary structure
    composition, and surface/buried composition. Values are computed using DSSP.
    Input: PDB or MMCIF filename
    Output: Dictionary of values obtained from dssp
    '''
    t = dssp_dataframe(filename, file_type)
    # print(t)
    sasa = calc_sasa(t)
    sstr = secondary_structure_summary(t)
    subu = calc_surface_buried(t)

    sasa.update(sstr)
    sasa.update(subu)

    return sasa


def get_ss_class(pdb_file, dssp_file, chain):
    """Define the secondary structure class of a PDB file at the specific chain

    Args:
        pdb_file:
        dssp_file:
        chain:

    Returns:

    """
    prag = pr.parsePDB(pdb_file)
    pr.parseDSSP(dssp_file, prag)
    alpha, threeTen, beta = get_dssp_ss_content_multiplechains(prag, chain)

    if alpha == 0 and beta > 0:
        classification = 'all-beta'
    elif beta == 0 and alpha > 0:
        classification = 'all-alpha'
    elif beta == 0 and alpha == 0:
        classification = 'mixed'
    elif float(alpha) / beta >= 20:
        classification = 'all-alpha'
    else:
        classification = 'mixed'

    return classification


if __name__ == '__main__':
    import glob
    files = glob.glob('../test_files/structures/*')
    # print(files)
    for f in files:
        print(f)
        # print(all_dssp_props(f))



import json
import requests
import tempfile
import ssbio.databases.pdb
import ssbio.utils
import os.path as op
from bioservices.uniprot import UniProt
import ssbio.databases.uniprot
bs_unip = UniProt()


def get_pdbs_for_gene(bigg_model, bigg_gene, cache_dir=tempfile.gettempdir()):
    """Attempt to get a rank-ordered list of available PDB structures for a BiGG Model and its gene.

    Args:
        bigg_model: BiGG Model ID
        bigg_gene: BiGG Gene ID

    Returns:
        list: rank-ordered list of tuples of (pdb_id, chain_id)

    """
    my_structures = []

    # Download gene info
    gene = ssbio.utils.request_json(link='http://bigg.ucsd.edu/api/v2/models/{}/genes/{}'.format(bigg_model, bigg_gene),
                                    outfile='{}_{}.json'.format(bigg_model, bigg_gene),
                                    outdir=cache_dir,
                                    force_rerun_flag=False)

    uniprots = []
    if 'database_links' in gene:
        if 'UniProt' in gene['database_links']:
            uniprots = [x['id'] for x in gene['database_links']['UniProt']]
        elif 'NCBI GI' in gene['database_links']:
            uniprots = []
            gis = [x['id'] for x in gene['database_links']['NCBI GI']]
            gi_uniprots = bs_unip.mapping(fr='P_GI', to='ACC', query=gis).values()
            uniprots.extend(gi_uniprots)
            uniprots = ssbio.utils.flatlist_dropdup(uniprots)
            uniprots = [x for x in uniprots if ssbio.databases.uniprot.is_valid_uniprot_id(x)]

    if uniprots:
        for u in uniprots:
            get_best_structure = ssbio.databases.pdb.best_structures(uniprot_id=u, outdir=cache_dir)
            if get_best_structure:
                for best_structure in get_best_structure:
                    my_structures.append((best_structure['pdb_id'], best_structure['chain_id']))

    return my_structures



import logging

from Bio.PDB.Polypeptide import one_to_three
from Bio.SeqUtils.ProtParam import ProteinAnalysis

import ssbio.protein.sequence.utils
import ssbio.utils

log = logging.getLogger(__name__)


# TODO: function to predict secondary structure (PSIPRED)
# TODO: function to predict solvent accessibility/resdepth just based on sequence (SCRATCH, RDPred)
# See more: https://www.researchgate.net/publication/235633540_Recent_Advances_in_Predicting_Functional_Impact_of_Single_Amino_Acid_Polymorphisms_A_Review_of_Useful_Features_Computational_Methods_and_Available_Tools

_aa_property_dict_one = {
    'Aliphatic': ['A', 'I', 'L', 'V'],
    'Aromatic' : ['F', 'H', 'W', 'Y'],
    'Non-polar': ['A', 'C', 'F', 'G', 'I', 'L', 'M', 'P', 'V', 'W', 'Y'],
    'Polar'    : ['D', 'E', 'H', 'K', 'N', 'Q', 'R', 'S', 'T'],
    'Charged'  : ['D', 'E', 'H', 'K', 'R'],
    'Basic'    : ['H', 'K', 'R'],
    'Acidic'   : ['D', 'E']}
# 'Tiny': ['A','C','G','S','T']
# 'Small': ['A','C','D','G','N','P','S','T','V']

_aa_property_dict_three = {k: [one_to_three(x) for x in v] for k, v in _aa_property_dict_one.items()}


def biopython_protein_analysis(inseq):
    """Utiize Biopython's ProteinAnalysis module to return general sequence properties of an amino acid string.

    For full definitions see: http://biopython.org/DIST/docs/api/Bio.SeqUtils.ProtParam.ProteinAnalysis-class.html

    Args:
        inseq: Amino acid sequence

    Returns:
        dict: Dictionary of sequence properties. Some definitions include:
        instability_index: Any value above 40 means the protein is unstable (has a short half life).
        secondary_structure_fraction: Percentage of protein in helix, turn or sheet

    TODO:
        Finish definitions of dictionary

    """

    inseq = ssbio.protein.sequence.utils.cast_to_str(inseq)

    analysed_seq = ProteinAnalysis(inseq)

    info_dict = {}
    # info_dict['amino_acids_content'] = analysed_seq.count_amino_acids()
    # info_dict['amino_acids_percent'] = analysed_seq.get_amino_acids_percent()
    # info_dict['length'] = analysed_seq.length
    info_dict['monoisotopic'] = analysed_seq.monoisotopic
    info_dict['molecular_weight'] = analysed_seq.molecular_weight()
    info_dict['aromaticity'] = analysed_seq.aromaticity()
    info_dict['instability_index'] = analysed_seq.instability_index()
    # TODO: What is flexibility?
    # info_dict['flexibility'] = analysed_seq.flexibility()
    info_dict['isoelectric_point'] = analysed_seq.isoelectric_point()

    # Separated secondary_structure_fraction into each definition
    # info_dict['secondary_structure_fraction'] = analysed_seq.secondary_structure_fraction()
    info_dict['percent_helix_naive'] = analysed_seq.secondary_structure_fraction()[0]
    info_dict['percent_turn_naive'] = analysed_seq.secondary_structure_fraction()[1]
    info_dict['percent_strand_naive'] = analysed_seq.secondary_structure_fraction()[2]

    return info_dict


def emboss_pepstats_on_fasta(infile, outfile='', outdir='', outext='.pepstats', force_rerun=False):
    """Run EMBOSS pepstats on a FASTA file.

    Args:
        infile: Path to FASTA file
        outfile: Name of output file without extension
        outdir: Path to output directory
        outext: Extension of results file, default is ".pepstats"
        force_rerun: Flag to rerun pepstats

    Returns:
        str: Path to output file.

    """

    # Create the output file name
    outfile = ssbio.utils.outfile_maker(inname=infile, outname=outfile, outdir=outdir, outext=outext)

    # Run pepstats
    program = 'pepstats'
    pepstats_args = '-sequence="{}" -outfile="{}"'.format(infile, outfile)
    cmd_string = '{} {}'.format(program, pepstats_args)
    ssbio.utils.command_runner(cmd_string, force_rerun_flag=force_rerun, outfile=outfile, silent=True)

    return outfile


def emboss_pepstats_parser(infile):
    """Get dictionary of pepstats results.

    Args:
        infile: Path to pepstats outfile

    Returns:
        dict: Parsed information from pepstats

    TODO:
        Only currently parsing the bottom of the file for percentages of properties.

    """
    with open(infile) as f:
        lines = f.read().split('\n')

    info_dict = {}

    for l in lines[38:47]:
        info = l.split('\t')
        cleaninfo = list(filter(lambda x: x != '', info))
        prop = cleaninfo[0]
        num = cleaninfo[2]
        percent = float(cleaninfo[-1]) / float(100)

        info_dict['percent_' + prop.lower()] = percent

    return info_dict


def residue_biochemical_definition(res):
    # TODO: docstring
    resprop = []
    for k, v in _aa_property_dict_one.items():
        if res in v:
            resprop.append(k)

    return resprop


def grantham_score(ref_aa, mut_aa):
    """https://github.com/ashutoshkpandey/Annotation/blob/master/Grantham_score_calculator.py"""
    grantham = {
        'S': {'R': 110, 'L': 145, 'P': 74, 'T': 58, 'A': 99, 'V': 124, 'G': 56, 'I': 142, 'F': 155, 'Y': 144, 'C': 112,
              'H': 89, 'Q': 68, 'N': 46, 'K': 121, 'D': 65, 'E': 80, 'M': 135, 'W': 177},
        'R': {'R': 0, 'L': 102, 'P': 103, 'T': 71, 'A': 112, 'V': 96, 'G': 125, 'I': 97, 'F': 97, 'Y': 77, 'C': 180,
              'H': 29, 'Q': 43, 'N': 86, 'K': 26, 'D': 96, 'E': 54, 'M': 91, 'W': 101, 'S': 0},
        'L': {'R': 0, 'L': 0, 'P': 98, 'T': 92, 'A': 96, 'V': 32, 'G': 138, 'I': 5, 'F': 22, 'Y': 36, 'C': 198, 'H': 99,
              'Q': 113, 'N': 153, 'K': 107, 'D': 172, 'E': 138, 'M': 15, 'W': 61, 'S': 0},
        'P': {'R': 0, 'L': 0, 'P': 0, 'T': 38, 'A': 27, 'V': 68, 'G': 42, 'I': 95, 'F': 114, 'Y': 110, 'C': 169,
              'H': 77, 'Q': 76, 'N': 91, 'K': 103, 'D': 108, 'E': 93, 'M': 87, 'W': 147, 'S': 0},
        'T': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 58, 'V': 69, 'G': 59, 'I': 89, 'F': 103, 'Y': 92, 'C': 149, 'H': 47,
              'Q': 42, 'N': 65, 'K': 78, 'D': 85, 'E': 65, 'M': 81, 'W': 128, 'S': 0},
        'A': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 64, 'G': 60, 'I': 94, 'F': 113, 'Y': 112, 'C': 195, 'H': 86,
              'Q': 91, 'N': 111, 'K': 106, 'D': 126, 'E': 107, 'M': 84, 'W': 148, 'S': 0},
        'V': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 109, 'I': 29, 'F': 50, 'Y': 55, 'C': 192, 'H': 84,
              'Q': 96, 'N': 133, 'K': 97, 'D': 152, 'E': 121, 'M': 21, 'W': 88, 'S': 0},
        'G': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 135, 'F': 153, 'Y': 147, 'C': 159, 'H': 98,
              'Q': 87, 'N': 80, 'K': 127, 'D': 94, 'E': 98, 'M': 127, 'W': 184, 'S': 0},
        'I': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 21, 'Y': 33, 'C': 198, 'H': 94,
              'Q': 109, 'N': 149, 'K': 102, 'D': 168, 'E': 134, 'M': 10, 'W': 61, 'S': 0},
        'F': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 22, 'C': 205, 'H': 100,
              'Q': 116, 'N': 158, 'K': 102, 'D': 177, 'E': 140, 'M': 28, 'W': 40, 'S': 0},
        'Y': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 194, 'H': 83,
              'Q': 99, 'N': 143, 'K': 85, 'D': 160, 'E': 122, 'M': 36, 'W': 37, 'S': 0},
        'C': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 174,
              'Q': 154, 'N': 139, 'K': 202, 'D': 154, 'E': 170, 'M': 196, 'W': 215, 'S': 0},
        'H': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 24,
              'N': 68, 'K': 32, 'D': 81, 'E': 40, 'M': 87, 'W': 115, 'S': 0},
        'Q': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 46, 'K': 53, 'D': 61, 'E': 29, 'M': 101, 'W': 130, 'S': 0},
        'N': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 94, 'D': 23, 'E': 42, 'M': 142, 'W': 174, 'S': 0},
        'K': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 0, 'D': 101, 'E': 56, 'M': 95, 'W': 110, 'S': 0},
        'D': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 0, 'D': 0, 'E': 45, 'M': 160, 'W': 181, 'S': 0},
        'E': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 0, 'D': 0, 'E': 0, 'M': 126, 'W': 152, 'S': 0},
        'M': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 0, 'D': 0, 'E': 0, 'M': 0, 'W': 67, 'S': 0},
        'W': {'R': 0, 'L': 0, 'P': 0, 'T': 0, 'A': 0, 'V': 0, 'G': 0, 'I': 0, 'F': 0, 'Y': 0, 'C': 0, 'H': 0, 'Q': 0,
              'N': 0, 'K': 0, 'D': 0, 'E': 0, 'M': 0, 'W': 0, 'S': 0}}

    score = 0

    if ref_aa not in grantham or mut_aa not in grantham:
        log.error('{} to {}: a residue is not in the Grantham matrix'.format(ref_aa, mut_aa))
        return score, 'Unknown'

    if ref_aa == mut_aa:
        return score, 'Conservative'
    else:
        if int(grantham[ref_aa][mut_aa]) != 0:
            score += int(grantham[ref_aa][mut_aa])
        else:
            score += int(grantham[mut_aa][ref_aa])

    if score > 150:
        return score, "Radical"
    elif 150 >= score > 100:
        return score, "Moderately Radical"
    elif 100 >= score > 50:
        return score, "Moderately Conservative"
    else:
        return score, "Conservative"


def flexibility_index(aa_one):
    """From Smith DK, Radivoja P, ObradovicZ, et al. Improved amino acid flexibility parameters, Protein Sci.2003, 12:1060

    Author: Ke Chen

    Args:
        aa_one:

    Returns:

    """
    aa_flexibility_index={}
    aa_flexibility_index['A']=-0.605
    aa_flexibility_index['C']=-0.692
    aa_flexibility_index['D']=-0.279
    aa_flexibility_index['E']=-0.160
    aa_flexibility_index['F']=-0.719
    aa_flexibility_index['G']=-0.537
    aa_flexibility_index['H']=-0.662
    aa_flexibility_index['I']=-0.682
    aa_flexibility_index['K']=-0.043
    aa_flexibility_index['L']=-0.631
    aa_flexibility_index['M']=-0.626
    aa_flexibility_index['N']=-0.381
    aa_flexibility_index['P']=-0.271
    aa_flexibility_index['Q']=-0.368
    aa_flexibility_index['R']=-0.448
    aa_flexibility_index['S']=-0.424
    aa_flexibility_index['T']=-0.525
    aa_flexibility_index['V']=-0.669
    aa_flexibility_index['W']=-0.727
    aa_flexibility_index['Y']=-0.721
    return aa_flexibility_index[aa_one]


import os.path as op
import re
import warnings

import bioservices
import pandas as pd
import requests
from dateutil.parser import parse as dateparse

import ssbio.utils
from ssbio.protein.sequence.seqprop import SeqProp

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from six.moves.urllib.request import urlretrieve

import logging
log = logging.getLogger(__name__)


# import cachetools
# SEVEN_DAYS = 60 * 60 * 24 * 7

bsup = bioservices.uniprot.UniProt()


class UniProtProp(SeqProp):
    """Class to parse through UniProt metadata all at once
    """

    def __init__(self, uniprot_acc, sequence_path=None, metadata_path=None):
        SeqProp.__init__(self, ident=uniprot_acc, sequence_path=sequence_path, metadata_path=metadata_path)
        self.ec_number = None
        self.entry_version = None
        self.pfam = None
        self.reviewed = False
        self.seq_version = None

        if uniprot_acc:
            if not is_valid_uniprot_id(uniprot_acc):
                raise ValueError("Invalid UniProt ID!")
            self.uniprot = uniprot_acc

        if metadata_path:
            self.load_metadata_path(metadata_path)

    def load_metadata_path(self, metadata_file):
        SeqProp.load_metadata_path(self, metadata_file)
        self.update(parse_uniprot_txt_file(metadata_file), overwrite=True,
                    only_keys=['description', 'kegg', 'refseq', 'ec_number', 'entry_version', 'gene_name',
                               'pfam', 'pdbs', 'reviewed', 'seq_version'])

    def download_seq_file(self, outdir, force_rerun=False):
        """Download and load the UniProt sequence file"""
        uniprot_seq_file = download_uniprot_file(uniprot_id=self.id,
                                                 filetype='fasta',
                                                 outdir=outdir,
                                                 force_rerun=force_rerun)

        self.load_sequence_path(uniprot_seq_file)

    def download_metadata_file(self, outdir, force_rerun=False):
        """Download and load the UniProt sequence file"""
        uniprot_metadata_file = download_uniprot_file(uniprot_id=self.id,
                                                      filetype='txt',
                                                      outdir=outdir,
                                                      force_rerun=force_rerun)

        self.load_metadata_path(uniprot_metadata_file)

    def ranking_score(self):
        """Provide a score for this UniProt ID based on reviewed (True=1, False=0) + number of PDBs

        Returns:
            int: Scoring for this ID

        """
        return self.reviewed + self.num_pdbs


def is_valid_uniprot_id(instring):
    """Check if a string is a valid UniProt ID.

    See regex from: http://www.uniprot.org/help/accession_numbers

    Args:
        instring: any string identifier

    Returns: True if the string is a valid UniProt ID

    """
    valid_id = re.compile("[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2}")
    if valid_id.match(str(instring)):
        return True
    else:
        return False


# TODO: method to blast UniProt to find a 100% sequence match
def blast_uniprot(seq_str, seq_ident=1, evalue=0.0001, reviewed_only=True, organism=None):
    """BLAST the UniProt db to find what IDs match the sequence input

    Args:
        seq_str: Sequence string
        seq_ident: Percent identity to match
        evalue: E-value of BLAST hit

    Returns:


    """
    pass


def get_fasta(uniprot_id):
    """Get the protein sequence for a UniProt ID as a string.

    Args:
        uniprot_id: Valid UniProt ID

    Returns:
        str: String of the protein (amino acid) sequence

    """
    if not is_valid_uniprot_id(uniprot_id):
        raise ValueError("Invalid UniProt ID!")

    # Silencing the "Will be moved to Biokit" message
    with ssbio.utils.suppress_stdout():
        return bsup.get_fasta_sequence(uniprot_id)


def uniprot_reviewed_checker(uniprot_id):
    """Check if a single UniProt ID is reviewed or not.

    Args:
        uniprot_id:

    Returns:
        bool: If the entry is reviewed

    """
    if not is_valid_uniprot_id(uniprot_id):
        raise ValueError("Invalid UniProt ID!")

    query_string = 'id:' + uniprot_id

    uni_rev_raw = StringIO(bsup.search(query_string, columns='id,reviewed', frmt='tab'))
    uni_rev_df = pd.read_table(uni_rev_raw, sep='\t', index_col=0)
    uni_rev_df = uni_rev_df.fillna(False)
    uni_rev_df = uni_rev_df[pd.notnull(uni_rev_df.Status)]

    uni_rev_df = uni_rev_df.replace(to_replace="reviewed", value=True)
    uni_rev_df = uni_rev_df.replace(to_replace="unreviewed", value=False)
    uni_rev_dict_adder = uni_rev_df.to_dict()['Status']

    return uni_rev_dict_adder[uniprot_id]


def uniprot_reviewed_checker_batch(uniprot_ids):
    """Batch check if uniprot IDs are reviewed or not

    Args:
        uniprot_ids: UniProt ID or list of UniProt IDs

    Returns:
        A dictionary of {UniProtID: Boolean}

    """
    uniprot_ids = ssbio.utils.force_list(uniprot_ids)

    invalid_ids = [i for i in uniprot_ids if not is_valid_uniprot_id(i)]
    uniprot_ids = [i for i in uniprot_ids if is_valid_uniprot_id(i)]

    if invalid_ids:
        warnings.warn("Invalid UniProt IDs {} will be ignored".format(invalid_ids))

    # splitting query up into managable sizes (200 IDs each)
    Nmax = 200
    N, rest = divmod(len(uniprot_ids), Nmax)

    uni_rev_dict = {}

    if rest > 0:
        N += 1
    for i in range(0, N):
        i1 = i * Nmax
        i2 = (i + 1) * Nmax
        if i2 > len(uniprot_ids):
            i2 = len(uniprot_ids)

        query = uniprot_ids[i1:i2]

        query_string = ''
        for x in query:
            query_string += 'id:' + x + '+OR+'
        query_string = query_string.strip('+OR+')

        uni_rev_raw = StringIO(bsup.search(query_string, columns='id,reviewed', frmt='tab'))
        uni_rev_df = pd.read_table(uni_rev_raw, sep='\t', index_col=0)
        uni_rev_df = uni_rev_df.fillna(False)

        # no_metadata = uni_rev_df[pd.isnull(uni_rev_df.Status)].index.tolist()
        # if no_metadata:
        #     warnings.warn("Unable to retrieve metadata for {}.".format(no_metadata))
        uni_rev_df = uni_rev_df[pd.notnull(uni_rev_df.Status)]

        uni_rev_df = uni_rev_df.replace(to_replace="reviewed", value=True)
        uni_rev_df = uni_rev_df.replace(to_replace="unreviewed", value=False)
        uni_rev_dict_adder = uni_rev_df.to_dict()['Status']
        uni_rev_dict.update(uni_rev_dict_adder)

    return uni_rev_dict


def uniprot_ec(uniprot_id):
    """Retrieve the EC number annotation for a UniProt ID.

    Args:
        uniprot_id: Valid UniProt ID

    Returns:

    """
    if not is_valid_uniprot_id(uniprot_id):
        raise ValueError('Invalid UniProt ID')

    r = requests.post('http://www.uniprot.org/uniprot/?query=%s&columns=ec&format=tab' % uniprot_id)
    ec = r.content.decode('utf-8').splitlines()[1]

    if len(ec) == 0:
        ec = None

    return ec


def uniprot_sites(uniprot_id):
    """Retrive a dictionary of UniProt sites

    Sites are defined here: http://www.uniprot.org/help/site and here: http://www.uniprot.org/help/function_section

    Args:
        uniprot_id: Valid UniProt ID

    Returns:

    """
    if not is_valid_uniprot_id(uniprot_id):
        raise ValueError('Invalid UniProt ID')

    r = requests.post('http://www.uniprot.org/uniprot/%s.gff' % uniprot_id)
    gff = StringIO(r.content.decode('utf-8'))

    try:
        gff_df = pd.read_table(gff, sep='\t', skiprows=2, header=None)
    except ValueError:
        return pd.DataFrame()

    gff_df.drop([0, 1, 5, 6, 7, 9], axis=1, inplace=True)
    gff_df.columns = ['type', 'seq_start', 'seq_end', 'notes']

    return gff_df


def download_uniprot_file(uniprot_id, filetype, outdir='', force_rerun=False):
    """Download a UniProt file for a UniProt ID/ACC

    Args:
        uniprot_id: Valid UniProt ID
        filetype: txt, fasta, xml, rdf, or gff
        outdir: Directory to download the file

    Returns:
        str: Absolute path to file

    """
    if not is_valid_uniprot_id(uniprot_id):
        raise ValueError('Invalid UniProt ID')

    my_file = '{}.{}'.format(uniprot_id, filetype)
    url = 'http://www.uniprot.org/uniprot/{}'.format(my_file)
    outfile = op.join(outdir, my_file)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        urlretrieve(url, outfile)

    return outfile


def parse_uniprot_txt_file(infile):
    """Parse a raw UniProt metadata file and return a dictionary.

    Args:
        infile: Path to metadata file

    Returns:
        dict: Metadata dictionaruy

    """
    uniprot_metadata_dict = {}

    metadata = old_parse_uniprot_txt_file(infile)
    metadata_key = list(metadata.keys())[0]

    uniprot_metadata_dict['seq_len'] = len(str(metadata[metadata_key]['sequence']))
    uniprot_metadata_dict['reviewed'] = metadata[metadata_key]['is_reviewed']
    uniprot_metadata_dict['seq_version'] = metadata[metadata_key]['sequence_version']
    uniprot_metadata_dict['entry_version'] = metadata[metadata_key]['entry_version']
    if 'gene' in metadata[metadata_key]:
        uniprot_metadata_dict['gene_name'] = metadata[metadata_key]['gene']
    if 'description' in metadata[metadata_key]:
        uniprot_metadata_dict['description'] = metadata[metadata_key]['description']
    if 'refseq' in metadata[metadata_key]:
        uniprot_metadata_dict['refseq'] = metadata[metadata_key]['refseq']
    if 'kegg' in metadata[metadata_key]:
        uniprot_metadata_dict['kegg'] = metadata[metadata_key]['kegg']
    if 'ec' in metadata[metadata_key]:
        uniprot_metadata_dict['ec_number'] = metadata[metadata_key]['ec']
    if 'pfam' in metadata[metadata_key]:
        uniprot_metadata_dict['pfam'] = metadata[metadata_key]['pfam']
    if 'pdbs' in metadata[metadata_key]:
        uniprot_metadata_dict['pdbs'] = list(set(metadata[metadata_key]['pdbs']))
    return uniprot_metadata_dict


def old_parse_uniprot_txt_file(infile):
    """From: boscoh/uniprot github
    Parses the text of metadata retrieved from uniprot.org.

    Only a few fields have been parsed, but this provides a
    template for the other fields.

    A single description is generated from joining alternative
    descriptions.

    Returns a dictionary with the main UNIPROT ACC as keys.
    """
    with open(infile, 'r') as txt:
        cache_txt = txt.read()
        tag = None
        uniprot_id = None
        metadata_by_seqid = {}
        for l in cache_txt.splitlines():
            test_tag = l[:5].strip()
            if test_tag and test_tag != tag:
                tag = test_tag
            line = l[5:].strip()
            words = line.split()
            if tag == "ID":
                uniprot_id = words[0]
                is_reviewed = words[1].startswith('Reviewed')
                length = int(words[2])
                metadata_by_seqid[uniprot_id] = {
                    'id': uniprot_id,
                    'is_reviewed': is_reviewed,
                    'length': length,
                    'sequence': '',
                    'accs': [],
                }
                entry = metadata_by_seqid[uniprot_id]
            if tag == "DT":
                # DT   01-OCT-1996, integrated into UniProtKB/Swiss-Prot.
                # DT   17-OCT-2006, sequence version 3.
                # DT   22-JUL-2015, entry version 166.
                comma_split = line.split(',')
                if 'sequence version' in comma_split[1]:
                    # print 'sequence_version', comma_split[0],
                    # dateparse(comma_split[0]).date()
                    entry['sequence_version'] = str(
                        dateparse(comma_split[0]).date())
                elif 'entry version' in comma_split[1]:
                    # print 'entry_version', comma_split[0],
                    # dateparse(comma_split[0]).date()
                    entry['entry_version'] = str(
                        dateparse(comma_split[0]).date())
            if tag == "SQ":
                if words[0] != "SEQUENCE":
                    entry['sequence'] += ''.join(words)
            if tag == "AC":
                accs = [w.replace(";", "") for w in words]
                entry['accs'].extend(accs)
            if tag == "DR":
                if 'PDB' in words[0]:
                    if 'pdb' not in entry:
                        entry['pdb'] = words[1][:-1]
                    if 'pdbs' not in entry:
                        entry['pdbs'] = []
                    entry['pdbs'].append(words[1][:-1])
                if 'RefSeq' in words[0]:
                    if 'refseq' not in entry:
                        entry['refseq'] = []
                    ids = [w[:-1] for w in words[1:]]
                    entry['refseq'].extend(ids)
                if 'KEGG' in words[0]:
                    if 'kegg' not in entry:
                        entry['kegg'] = []
                    ids = [w[:-1] for w in words[1:]]
                    ids = filter(lambda w: len(w) > 1, ids)
                    entry['kegg'].extend(ids)
                if 'GO' in words[0]:
                    if 'go' not in entry:
                        entry['go'] = []
                    entry['go'].append(' '.join(words[1:]))
                if 'Pfam' in words[0]:
                    if 'pfam' not in entry:
                        entry['pfam'] = []
                    entry['pfam'].append(words[1][:-1])
            if tag == "GN":
                if 'gene' not in entry and len(words) > 0:
                    pieces = words[0].split("=")
                    if len(pieces) > 1 and 'name' in pieces[0].lower():
                        entry['gene'] = pieces[1].replace(
                            ';', '').replace(',', '')
            if tag == "OS":
                # OS   Homo sapiens (Human).
                if 'organism' not in entry:
                    entry['organism'] = ""
                entry['organism'] += line
            if tag == "DE":
                if 'descriptions' not in entry:
                    entry['descriptions'] = []
                entry['descriptions'].append(line)
            if tag == "CC":
                if 'comment' not in entry:
                    entry['comment'] = ''
                entry['comment'] += line + '\n'

        for entry in metadata_by_seqid.values():
            descriptions = entry['descriptions']
            for i in reversed(range(len(descriptions))):
                description = descriptions[i]
                if 'Short' in description or 'Full' in description or 'EC=' in description:
                    if 'Short' in description or 'Full' in description:
                        j = description.find('=')
                        descriptions[i] = description[j + 1:].replace(';', '')
                        if 'description' not in entry:
                            entry['description'] = []
                        entry['description'].append(descriptions[i])
                    if 'EC=' in description:
                        j = description.find('=')
                        descriptions[i] = description[j + 1:].replace(';', '')
                        if '{' in descriptions[i]:
                            descriptions[i] = descriptions[i].split(' {')[0]
                        if 'ec' not in entry:
                            entry['ec'] = []
                        entry['ec'].append(descriptions[i])
                else:
                    del descriptions[i]

    return metadata_by_seqid


# def uniprot_metadata_batch(uniprot_ids, outdir=None):
#     '''
#     Input: UniProt ID or IDs
#     Output: dictionary of metadata associated with the UniProt IDs
#     '''
#     # TODO: not working as of 2016-05-08 - seems to be a limit on the url requests
#     counter = 1
#
#     # if isinstance(uniprot_ids, str):
#     #     uniprot_ids = [uniprot_ids]
#     #     single = True
#     # else:
#     #     single = False
#     uniprot_ids = utils.force_list(uniprot_ids)
#
#     uniprot_metadata_raw = bs_unip.retrieve(uniprot_ids, frmt='txt')
#     uniprot_metadata_final = {}
#
#     for uniprot_id in tqdm(uniprot_ids):
#
#         uniprot_metadata_dict = {}
#
#         if single:
#             metadata = parse_uniprot_txt_file(uniprot_metadata_raw)
#         else:
#             metadata = parse_uniprot_txt_file(
#                 uniprot_metadata_raw[uniprot_ids.index(uniprot_id)])
#
#         metadata_key = list(metadata.keys())[0]
#
#         uniprot_metadata_dict['u_uniprot_acc'] = uniprot_id
#         uniprot_metadata_dict['u_seq'] = metadata[metadata_key]['sequence']
#         uniprot_metadata_dict['u_seq_len'] = len(
#             str(metadata[metadata_key]['sequence']))
#         uniprot_metadata_dict['u_reviewed'] = metadata[
#             metadata_key]['is_reviewed']
#         uniprot_metadata_dict['u_seq_version'] = metadata[
#             metadata_key]['sequence_version']
#         uniprot_metadata_dict['u_entry_version'] = metadata[
#             metadata_key]['entry_version']
#         if 'gene' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_gene_name'] = metadata[
#                 metadata_key]['gene']
#         if 'description' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_description'] = metadata[
#                 metadata_key]['description']
#         if 'refseq' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_refseq'] = metadata[
#                 metadata_key]['refseq']
#         if 'kegg' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_kegg_id'] = metadata[
#                 metadata_key]['kegg']
#         if 'go' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_go'] = metadata[metadata_key]['go']
#         if 'ec' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_ec_number'] = metadata[
#                 metadata_key]['ec']
#         if 'pfam' in metadata[metadata_key]:
#             uniprot_metadata_dict['u_pfam'] = metadata[
#                 metadata_key]['pfam']
#
#         uniprot_metadata_final[uniprot_id] = uniprot_metadata_dict
#     return uniprot_metadata_final





import tempfile
import os.path as op
from requests import get  # to make GET request
import pandas as pd
import gzip
import urllib.request


def download(url, file_name):
    # open in binary mode
    with open(file_name, "wb") as file:
        # get request
        response = get(url)
        # write to file
        file.write(response.content)
    return file_name


def download_parse_uniprot_humsavar(wd):
    # TODO: this is not functional as of now - the whitespace separator doesn't work for the last column
    baseURL = "http://www.uniprot.org/docs/"
    filename = "humsavar.txt"

    downloaded_file = download(baseURL + filename, op.join(tempfile.gettempdir(), filename))
    parsed_file = pd.read_table(downloaded_file, skiprows=30, skipfooter=4, sep=r"\s*",
                                header=None, engine='python').rename(columns={0: 'u_gene_name', 1: 'u_uniprot_acc',
                                                                              2: 'u_ft_id', 3: 'u_var', 4: 'u_var_type',
                                                                              5: 'u_dbsnp', 6: 'u_disease'})
    outfile = op.join(wd, 'humsavar.csv')
    parsed_file.to_csv(outfile)
    return outfile


def download_parse_uniprot_variation(wd):
    baseURL = "ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/variants/"
    filename = "homo_sapiens_variation.txt.gz"

    req = urllib.request.urlopen(baseURL + filename)
    decompressedFile = gzip.GzipFile(fileobj=req)

    outfile = op.join(tempfile.gettempdir(), filename)
    with open(outfile, 'w') as f:
        f.write(decompressedFile.read())

    uniprot_snps_new_df = pd.read_table(outfile, skiprows=139, delimiter='\t', header=None).rename(
        columns={0: 'u_gene_name', 1: 'u_uniprot_acc', 2: 'u_aa_change', 3: 'u_dbsnp',
                 4: 'u_var_type', 5: 'u_consq_type', 6: 'u_clin_sig', 7: 'u_phenotype',
                 8: 'u_cytogenic_band', 9: 'u_chr_loc', 10: 'ensg', 11: 'enst', 12: 'ensp'})

    outfile_pandas = op.join(wd, 'homo_sapiens_variation.csv')
    uniprot_snps_new_df.to_csv(outfile_pandas)
    return outfile_pandas


if __name__ == "__main__":
    pass


import logging
import os.path as op
import subprocess
import tempfile
from collections import defaultdict
from itertools import count, groupby

import numpy as np
import pandas as pd
from Bio import AlignIO
from Bio import pairwise2
from Bio.Align import MultipleSeqAlignment
from Bio.SubsMat import MatrixInfo as matlist

import ssbio.protein.sequence.utils
import ssbio.utils

log = logging.getLogger(__name__)

# quiet the SettingWithCopyWarning when converting dtypes in get_deletions/mutations methods
pd.options.mode.chained_assignment = None

from Bio.PDB.Polypeptide import one_to_three

_aa_property_dict_one = {'Tiny': ['A','C','G','S','T'],
'Small': ['A','C','D','G','N','P','S','T','V'],
'Aliphatic': ['A','I','L','V'],
'Aromatic': ['F','H','W','Y'],
'Non-polar': ['A','C','F','G','I','L','M','P','V','W','Y'],
'Polar': ['D','E','H','K','N','Q','R','S','T'],
'Charged': ['D','E','H','K','R'],
'Basic': ['H','K','R'],
'Acidic': ['D','E']}

_aa_property_dict_three = {k: [one_to_three(x) for x in v] for k,v in _aa_property_dict_one.items()}


def pairwise_sequence_alignment(a_seq, b_seq, engine, a_seq_id=None, b_seq_id=None,
                                gapopen=10, gapextend=0.5,
                                outfile=None, outdir=None, force_rerun=False):
    """Run a global pairwise sequence alignment between two sequence strings.

    Args:
        a_seq:
        b_seq:
        engine (str): 'biopython' or 'needle'
        a_seq_id:
        b_seq_id:
        gapopen:
        gapextend:
        outfile:
        outdir:
        force_rerun:

    Returns:
        MultipleSeqAlignment: Biopython object to represent an alignment

    """
    engine = engine.lower()

    if engine not in ['biopython', 'needle']:
        raise ValueError('{}: invalid engine'.format(engine))

    if not a_seq_id:
        a_seq_id = 'a_seq'
    if not b_seq_id:
        b_seq_id = 'b_seq'

    a_seq = ssbio.protein.sequence.utils.cast_to_str(a_seq)
    b_seq = ssbio.protein.sequence.utils.cast_to_str(b_seq)

    if engine == 'biopython':
        # TODO: allow different matrices? needle uses blosum62 by default, how to change that?
        # TODO: how to define gap open/extend when using matrix in biopython global alignment?
        log.warning('Gap penalties not implemented in Biopython yet')

        blosum62 = matlist.blosum62
        alignments = pairwise2.align.globaldx(a_seq, b_seq, blosum62)  # TODO: add gap penalties
        best_alignment = alignments[0]

        a = ssbio.protein.sequence.utils.cast_to_seq_record(best_alignment[0], id=a_seq_id)
        b = ssbio.protein.sequence.utils.cast_to_seq_record(best_alignment[1], id=b_seq_id)
        alignment = MultipleSeqAlignment([a, b], annotations={"score": best_alignment[2],
                                                              "start": best_alignment[3],
                                                              "end"  : best_alignment[4]})
        alignment.annotations['percent_identity'] = get_percent_identity(best_alignment[0], best_alignment[1]) * 100

        return alignment

    if engine == 'needle':
        alignment_file = run_needle_alignment(seq_a=a_seq, seq_b=b_seq, gapopen=gapopen, gapextend=gapextend,
                                              outdir=outdir, outfile=outfile, force_rerun=force_rerun)
        log.debug('Needle alignment at {}'.format(alignment_file))

        # Use AlignIO to parse the needle alignment, alignments[0] is the first alignment (the only one in pairwise)
        alignments = list(AlignIO.parse(alignment_file, "emboss"))
        alignment = alignments[0]

        # Rename the sequence IDs
        alignment[0].id = a_seq_id
        alignment[1].id = b_seq_id

        # Add needle statistics as annotations in the alignment object
        stats = needle_statistics(alignment_file)
        alignment_ids = list(stats.keys())
        if len(alignment_ids) > 1:
            raise ValueError('Needle alignment file contains more than one pairwise alignment')
        needle_id = alignment_ids[0]
        alignment.annotations['percent_identity'] = stats[needle_id]['percent_identity']
        alignment.annotations['percent_similarity'] = stats[needle_id]['percent_similarity']
        alignment.annotations['percent_gaps'] = stats[needle_id]['percent_gaps']
        alignment.annotations['score'] = stats[needle_id]['score']

        return alignment


def run_needle_alignment(seq_a, seq_b, gapopen=10, gapextend=0.5,
                         outdir=None, outfile=None, force_rerun=False):
    """Run the needle alignment program for two strings and return the raw alignment result.

    More info:
    EMBOSS needle: http://www.bioinformatics.nl/cgi-bin/emboss/help/needle
    Biopython wrapper: http://biopython.org/DIST/docs/tutorial/Tutorial.html#htoc84
    Using strings as input: https://www.biostars.org/p/91124/

    Args:
        id_a: ID of reference sequence
        seq_a (str, Seq, SeqRecord): Reference sequence
        id_b: ID of sequence to be aligned
        seq_b (str, Seq, SeqRecord): String representation of sequence to be aligned
        gapopen: Gap open penalty is the score taken away when a gap is created
        gapextend: Gap extension penalty is added to the standard gap penalty for each base or residue in the gap
        outdir (str, optional): Path to output directory. Default is the current directory.
        outfile (str, optional): Name of output file. If not set, is {id_a}_{id_b}_align.txt
        force_rerun (bool): Default False, set to True if you want to rerun the alignment if outfile exists.

    Returns:
        str: Raw alignment result of the needle alignment in srspair format.

    """
    if not outdir:
        outdir = ''

    # TODO: rewrite using utils functions - does not report error if needle is not installed currently
    # TODO: rethink outdir/outfile, also if this should return the tempfile or just a file object or whatever
    if not outfile:
        outfile = op.join(tempfile.gettempdir(), 'temp_alignment.needle')
    else:
        outfile = op.join(outdir, outfile)

    # Check if the outfile already exists, if so just return the filename
    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        seq_a = ssbio.protein.sequence.utils.cast_to_str(seq_a)
        seq_b = ssbio.protein.sequence.utils.cast_to_str(seq_b)

        cmd = 'needle -outfile="{}" -asequence=asis::{} -bsequence=asis::{} -gapopen={} -gapextend={}'.format(outfile, seq_a, seq_b, gapopen, gapextend)
        command = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   shell=True)
        out, err = command.communicate()
        # needle_cline = NeedleCommandline(asequence="asis::"+seq_a, bsequence="asis::"+seq_b,
        #                                  gapopen=gapopen, gapextend=gapextend,
        #                                  outfile=outfile)
        # stdout, stderr = needle_cline()

    return outfile


def run_needle_alignment_on_files(id_a, faa_a, id_b, faa_b, gapopen=10, gapextend=0.5,
                                  outdir='', outfile='', force_rerun=False):
    """Run the needle alignment program for two fasta files and return the raw alignment result.

    More info:
    EMBOSS needle: http://www.bioinformatics.nl/cgi-bin/emboss/help/needle
    Biopython wrapper: http://biopython.org/DIST/docs/tutorial/Tutorial.html#htoc84

    Args:
        id_a: ID of reference sequence
        faa_a: File path to reference sequence
        id_b: ID of sequence to be aligned
        faa_b: File path to sequence to be aligned
        gapopen: Gap open penalty is the score taken away when a gap is created
        gapextend: Gap extension penalty is added to the standard gap penalty for each base or residue in the gap
        outdir (str, optional): Path to output directory. Default is the current directory.
        outfile (str, optional): Name of output file. If not set, is {id_a}_{id_b}_align.txt
        force_rerun (bool): Default False, set to True if you want to rerun the alignment if outfile exists.

    Returns:
        str: Raw alignment result of the needle alignment in srspair format.

    """

    # TODO: rewrite using utils functions so we can check for needle installation
    # # If you don't want to save the output file, just run the alignment and return the raw results
    # if not outfile and not outdir:
    #     needle_cline = NeedleCommandline(asequence=faa_a, bsequence=faa_b,
    #                                      gapopen=gapopen, gapextend=gapextend,
    #                                      stdout=True, auto=True)
    #     stdout, stderr = needle_cline()
    #     raw_alignment_text = stdout.decode('utf-8')

    # Make a default name if no outfile is set
    if not outfile:
        outfile = op.join(outdir, '{}_{}.needle'.format(id_a, id_b))
    else:
        outfile = op.join(outdir, outfile)

    # Check if the outfile already exists
    if op.exists(outfile) and not force_rerun:
        return outfile
    # If it doesn't exist, or force_rerun=True, run the alignment
    else:
        cmd = 'needle -outfile="{}" -asequence="{}" -bsequence="{}" -gapopen={} -gapextend={}'.format(outfile,
                                                                                                      faa_a,
                                                                                                      faa_b,
                                                                                                      gapopen,
                                                                                                      gapextend)
        command = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   shell=True)
        out, err = command.communicate()

    return outfile


def get_percent_identity(a_aln_seq, b_aln_seq):
    """Get the percent identity between two alignment strings"""

    if len(a_aln_seq) != len(b_aln_seq):
        raise ValueError('Sequence lengths not equal - was an alignment run?')

    count = 0
    gaps = 0
    for n in range(0, len(a_aln_seq)):
        if a_aln_seq[n] == b_aln_seq[n]:
            if a_aln_seq[n] != "-":
                count += 1
            else:
                gaps += 1

    return count / float((len(a_aln_seq) - gaps))


def get_alignment_df(a_aln_seq, b_aln_seq, a_seq_id=None, b_seq_id=None):
    """Summarize two alignment strings in a dataframe.

    Args:
        a_aln_seq (str): Aligned sequence string
        b_aln_seq (str): Aligned sequence string
        a_seq_id (str): Optional ID of a_seq
        b_seq_id (str): Optional ID of b_aln_seq

    Returns:
        DataFrame: a per-residue level annotation of the alignment

    """
    if len(a_aln_seq) != len(b_aln_seq):
        raise ValueError('Sequence lengths not equal - was an alignment run?')

    if not a_seq_id:
        a_seq_id = 'a_seq'
    if not b_seq_id:
        b_seq_id = 'b_seq'

    a_aln_seq = ssbio.protein.sequence.utils.cast_to_str(a_aln_seq)
    b_aln_seq = ssbio.protein.sequence.utils.cast_to_str(b_aln_seq)

    a_idx = 1
    b_idx = 1

    appender = []

    for i, (a,b) in enumerate(zip(a_aln_seq, b_aln_seq)):
        to_append = {}

        if a == b and a != '-' and b != '-':
            aa_flag = 'match'
        if a != b and a == '-' and b != '-':
            aa_flag = 'insertion'
        if a != b and a != '-' and b == '-':
            aa_flag = 'deletion'
        if a != b and a != '-' and b == 'X':
            aa_flag = 'unresolved'
        if a != b and b != '-' and a == 'X':
            aa_flag = 'unresolved'
        elif a != b and a != '-' and b != '-':
            aa_flag = 'mutation'

        to_append['id_a'] = a_seq_id
        to_append['id_b'] = b_seq_id
        to_append['type'] = aa_flag

        if aa_flag == 'match' or aa_flag == 'unresolved' or aa_flag == 'mutation':
            to_append['id_a_aa'] = a
            to_append['id_a_pos'] = int(a_idx)
            to_append['id_b_aa'] = b
            to_append['id_b_pos'] = int(b_idx)
            a_idx += 1
            b_idx += 1

        if aa_flag == 'deletion':
            to_append['id_a_aa'] = a
            to_append['id_a_pos'] = int(a_idx)
            a_idx += 1

        if aa_flag == 'insertion':
            to_append['id_b_aa'] = b
            to_append['id_b_pos'] = int(b_idx)
            b_idx += 1

        appender.append(to_append)

    cols = ['id_a', 'id_b', 'type', 'id_a_aa', 'id_a_pos', 'id_b_aa', 'id_b_pos']
    alignment_df = pd.DataFrame.from_records(appender, columns=cols)
    alignment_df = alignment_df.fillna(value=np.nan)

    return alignment_df


def get_alignment_df_from_file(alignment_file, a_seq_id=None, b_seq_id=None):
    """Get a Pandas DataFrame of the Needle alignment results. Contains all positions of the sequences.

    Args:
        alignment_file:
        a_seq_id: Optional specification of the ID of the reference sequence
        b_seq_id: Optional specification of the ID of the aligned sequence

    Returns:
        Pandas DataFrame: all positions in the alignment

    """
    alignments = list(AlignIO.parse(alignment_file, "emboss"))
    alignment_df = pd.DataFrame(columns=['id_a', 'id_b', 'type', 'id_a_aa', 'id_a_pos', 'id_b_aa', 'id_b_pos'])

    for alignment in alignments:
        if not a_seq_id:
            a_seq_id = list(alignment)[0].id
        a_seq = str(list(alignment)[0].seq)
        if not b_seq_id:
            b_seq_id = list(alignment)[1].id
        b_seq = str(list(alignment)[1].seq)

        df = get_alignment_df(a_seq, b_seq, a_seq_id, b_seq_id)
        alignment_df = alignment_df.append(df).reset_index(drop=True)

    return alignment_df


def get_mutations(aln_df):
    """Get a list of residue numbers (in the original sequence's numbering) that are mutated

    Args:
        aln_df (DataFrame): Alignment DataFrame
        just_resnums: If only the residue numbers should be returned, instead of a list of tuples of
            (original_residue, resnum, mutated_residue)

    Returns:
        list: Residue mutations

    """
    mutation_df = aln_df[aln_df['type'] == 'mutation']
    tuples = []
    if not mutation_df.empty:
        subset = mutation_df[['id_a_aa', 'id_a_pos', 'id_b_aa']]
        subset['id_a_pos'] = subset['id_a_pos'].astype(int)
        tuples = [tuple(x) for x in subset.values]
    return tuples


def get_unresolved(aln_df):
    """Get a list of residue numbers (in the original sequence's numbering) that are unresolved

    Args:
        aln_df (DataFrame): Alignment DataFrame

    Returns:
        list: Residue numbers that are mutated

    """
    unresolved_df = aln_df[aln_df['type'] == 'unresolved']
    unresolved = []
    if not unresolved_df.empty:
        unresolved_df['id_a_pos'] = unresolved_df['id_a_pos'].astype(int)
        unresolved = unresolved_df.id_a_pos.tolist()
    return unresolved


def get_deletions(aln_df):
    """Get a list of tuples indicating the first and last residues of a deletion region, as well as the length of the deletion.

    Examples:
        # Deletion of residues 1 to 4, length 4
        >>> test = {'id_a': {0: 'a', 1: 'a', 2: 'a', 3: 'a'}, 'id_a_aa': {0: 'M', 1: 'G', 2: 'I', 3: 'T'}, 'id_a_pos': {0: 1.0, 1: 2.0, 2: 3.0, 3: 4.0}, 'id_b': {0: 'b', 1: 'b', 2: 'b', 3: 'b'}, 'id_b_aa': {0: np.nan, 1: np.nan, 2: np.nan, 3: np.nan}, 'id_b_pos': {0: np.nan, 1: np.nan, 2: np.nan, 3: np.nan}, 'type': {0: 'deletion', 1: 'deletion', 2: 'deletion', 3: 'deletion'}}
        >>> my_alignment = pd.DataFrame.from_dict(test)
        >>> get_deletions(my_alignment)
        [((1.0, 4.0), 4)]

    Args:
        aln_df (DataFrame): Alignment DataFrame

    Returns:
        list: A list of tuples with the format ((deletion_start_resnum, deletion_end_resnum), deletion_length)

    """

    deletion_df = aln_df[aln_df['type'] == 'deletion']
    if not deletion_df.empty:
        deletion_df['id_a_pos'] = deletion_df['id_a_pos'].astype(int)
    deletions = []

    for k, g in groupby(deletion_df.index, key=lambda n, c=count(): n - next(c)):
        tmp = list(g)
        deletion_indices = (min(tmp), max(tmp))

        deletion_start_ix = deletion_indices[0]
        deletion_end_ix = deletion_indices[1]

        deletion_length = deletion_end_ix - deletion_start_ix + 1

        id_a_pos_deletion_start = aln_df.ix[deletion_start_ix].id_a_pos
        id_a_pos_deletion_end = aln_df.ix[deletion_end_ix].id_a_pos

        deletion_region = (id_a_pos_deletion_start, id_a_pos_deletion_end)

        # Logging where the insertion is
        log.debug('Deletion of length {} at residues {}'.format(deletion_length, deletion_region))

        to_append = (deletion_region, deletion_length)
        deletions.append(to_append)

    return deletions


def get_insertions(aln_df):
    """Get a list of tuples indicating the first and last residues of a insertion region, as well as the length of the insertion.

    If the first tuple is:
        (-1, 1) that means the insertion is at the beginning of the original protein
        (X, Inf) where X is the length of the original protein, that means the insertion is at the end of the protein

    Examples:
        # Insertion at beginning, length 3
        >>> test = {'id_a': {0: 'a', 1: 'a', 2: 'a', 3: 'a'}, 'id_a_aa': {0: np.nan, 1: np.nan, 2: np.nan, 3: 'M'}, 'id_a_pos': {0: np.nan, 1: np.nan, 2: np.nan, 3: 1.0}, 'id_b': {0: 'b', 1: 'b', 2: 'b', 3: 'b'}, 'id_b_aa': {0: 'M', 1: 'M', 2: 'L', 3: 'M'}, 'id_b_pos': {0: 1, 1: 2, 2: 3, 3: 4}, 'type': {0: 'insertion', 1: 'insertion', 2: 'insertion', 3: 'match'}}
        >>> my_alignment = pd.DataFrame.from_dict(test)
        >>> get_insertions(my_alignment)
        [((-1, 1.0), 3)]

    Args:
        aln_df (DataFrame): Alignment DataFrame

    Returns:
        list: A list of tuples with the format ((insertion_start_resnum, insertion_end_resnum), insertion_length)

    """

    insertion_df = aln_df[aln_df['type'] == 'insertion']
    # if not insertion_df.empty: # don't need to do this for insertions
    #     insertion_df['id_a_pos'] = insertion_df['id_a_pos'].astype(int)

    insertions = []

    for k, g in groupby(insertion_df.index, key=lambda n, c=count(): n - next(c)):
        tmp = list(g)
        insertion_indices = (min(tmp), max(tmp))

        insertion_start = insertion_indices[0] - 1
        insertion_end = insertion_indices[1] + 1

        # Checking if insertion is at the beginning or end
        if insertion_start < 0:
            insertion_start = insertion_indices[0]
            insertion_length = insertion_end - insertion_start
        elif insertion_end >= len(aln_df):
            insertion_end = insertion_indices[1]
            insertion_length = insertion_end - insertion_start
        else:
            insertion_length = insertion_end - insertion_start - 1

        id_a_pos_insertion_start = aln_df.ix[insertion_start].id_a_pos
        id_a_pos_insertion_end = aln_df.ix[insertion_end].id_a_pos

        # Checking if insertion is at the beginning or end
        if np.isnan(id_a_pos_insertion_start) and id_a_pos_insertion_end == 1:
            insertion_region = (-1, id_a_pos_insertion_end)
        elif np.isnan(id_a_pos_insertion_end):
            insertion_region = (id_a_pos_insertion_start, float('Inf'))
        else:
            insertion_region = (id_a_pos_insertion_start, id_a_pos_insertion_end)

        # Logging where the insertion is
        if insertion_region[0] == -1:
            log.debug('Insertion of length {} at beginning'.format(insertion_length))
        elif insertion_region[1] == float('Inf'):
            log.debug('Insertion of length {} at end'.format(insertion_length))
        else:
            log.debug('Insertion of length {} at residues {}'.format(insertion_length, insertion_region))

        to_append = (insertion_region, insertion_length)
        insertions.append(to_append)

    return insertions


def map_resnum_a_to_resnum_b(a_resnum, a_aln, b_aln):
    """Map a residue number in a sequence to the corresponding residue number in an aligned sequence.

    Examples:
    >>> map_resnum_a_to_resnum_b(5, '--ABCDEF', 'XXABCDEF')
    7

    Args:
        a_resnum (int): Residue number in the first aligned sequence
        a_aln (str, Seq, SeqRecord): Aligned sequence string
        b_aln (str, Seq, SeqRecord): Aligned sequence string

    Returns:
        int: Residue number in the second aligned sequence
    """

    aln_df = get_alignment_df(a_aln, b_aln)
    maps = aln_df[aln_df.id_a_pos == a_resnum].id_b_pos.values

    if len(maps) > 1:
        print(maps)
        raise LookupError('More than one mapping for position {}'.format(a_resnum))

    b_resnum = maps[0]
    if np.isnan(b_resnum):
        return None

    return int(b_resnum)


def needle_statistics(infile):
    """Reads in a needle alignment file and spits out statistics of the alignment.

    Args:
        infile (str): Alignment file name

    Returns:
        dict: alignment_properties - a dictionary telling you the number of gaps, identity, etc.

    """

    alignments = list(AlignIO.parse(infile, "emboss"))
    alignment_properties = defaultdict(dict)

    with open(infile) as f:
        line = f.readline()

        for i in range(len(alignments)):
            while line.rstrip() != "#=======================================":
                line = f.readline()
                if not line:
                    raise StopIteration

            while line[0] == "#":
                # Read in the rest of this alignment header,
                # try and discover the number of records expected and their length
                parts = line[1:].split(":", 1)
                key = parts[0].lower().strip()
                if key == '1':
                    a_id = parts[1].strip()
                if key == '2':
                    b_id = parts[1].strip()
                if key == 'identity':
                    ident_parse = parts[1].strip().replace('(','').replace(')','').replace('%','').split()
                    ident_num = int(ident_parse[0].split('/')[0])
                    ident_percent = float(ident_parse[1])
                    alignment_properties[a_id + '_' + b_id]['identity'] = ident_num
                    alignment_properties[a_id + '_' + b_id]['percent_identity'] = ident_percent
                if key == 'similarity':
                    sim_parse = parts[1].strip().replace('(','').replace(')','').replace('%','').split()
                    sim_num = int(sim_parse[0].split('/')[0])
                    sim_percent = float(sim_parse[1])
                    alignment_properties[a_id + '_' + b_id]['similarity'] = sim_num
                    alignment_properties[a_id + '_' + b_id]['percent_similarity'] = sim_percent
                if key == 'gaps':
                    gap_parse = parts[1].strip().replace('(','').replace(')','').replace('%','').split()
                    gap_num = int(gap_parse[0].split('/')[0])
                    gap_percent = float(gap_parse[1])
                    alignment_properties[a_id + '_' + b_id]['gaps'] = gap_num
                    alignment_properties[a_id + '_' + b_id]['percent_gaps'] = gap_percent
                if key == 'score':
                    score = float(parts[1].strip())
                    alignment_properties[a_id + '_' + b_id]['score'] = score

                # And read in another line...
                line = f.readline()

    return alignment_properties